@@ -41,6 +41,7 @@
 
 #include <ev.h>
 
+#define NGHTTP2_NO_SSIZE_T
 #include <nghttp2/nghttp2.h>
 
 #include "http2.h"
@@ -69,6 +70,7 @@ struct Config {
   size_t num_worker;
   size_t max_concurrent_streams;
   ssize_t header_table_size;
+  ssize_t encoder_header_table_size;
   int window_bits;
   int connection_window_bits;
   uint16_t port;
@@ -81,6 +83,8 @@ struct Config {
   bool hexdump;
   bool echo_upload;
   bool no_content_length;
+  bool ktls;
+  bool no_rfc7540_pri;
   Config();
   ~Config();
 };
@@ -89,7 +93,8 @@ class Http2Handler;
 
 struct FileEntry {
   FileEntry(std::string path, int64_t length, int64_t mtime, int fd,
-            const std::string *content_type, ev_tstamp last_valid,
+            const std::string *content_type,
+            const std::chrono::steady_clock::time_point &last_valid,
             bool stale = false)
       : path(std::move(path)),
         length(length),
@@ -105,7 +110,7 @@ struct FileEntry {
   std::multimap<std::string, std::unique_ptr<FileEntry>>::iterator it;
   int64_t length;
   int64_t mtime;
-  ev_tstamp last_valid;
+  std::chrono::steady_clock::time_point last_valid;
   const std::string *content_type;
   FileEntry *dlnext, *dlprev;
   int fd;
@@ -163,19 +168,19 @@ public:
   int on_read();
   int on_write();
   int connection_made();
-  int verify_npn_result();
+  int verify_alpn_result();
 
   int submit_file_response(const StringRef &status, Stream *stream,
                            time_t last_modified, off_t file_length,
                            const std::string *content_type,
-                           nghttp2_data_provider *data_prd);
+                           nghttp2_data_provider2 *data_prd);
 
   int submit_response(const StringRef &status, int32_t stream_id,
-                      nghttp2_data_provider *data_prd);
+                      nghttp2_data_provider2 *data_prd);
 
   int submit_response(const StringRef &status, int32_t stream_id,
                       const HeaderRefs &headers,
-                      nghttp2_data_provider *data_prd);
+                      nghttp2_data_provider2 *data_prd);
 
   int submit_non_final_response(const std::string &status, int32_t stream_id);
 
@@ -241,7 +246,7 @@ private:
 };
 
 ssize_t file_read_callback(nghttp2_session *session, int32_t stream_id,
-                           uint8_t *buf, size_t length, int *eof,
+                           uint8_t *buf, size_t length, uint32_t *data_flags,
                            nghttp2_data_source *source, void *user_data);
 
 } // namespace nghttp2
