@@ -124,7 +124,7 @@ int Http3Session::on_write() { return -1; }
 void Http3Session::terminate() {}
 
 size_t Http3Session::max_concurrent_streams() {
-  return (size_t)client_->worker->config->max_concurrent_streams;
+  return client_->worker->config->max_concurrent_streams;
 }
 
 namespace {
@@ -147,6 +147,23 @@ int Http3Session::stream_close(int64_t stream_id, uint64_t app_error_code) {
   return 0;
 }
 
+namespace {
+int end_stream(nghttp3_conn *conn, int64_t stream_id, void *user_data,
+               void *stream_user_data) {
+  auto s = static_cast<Http3Session *>(user_data);
+  if (s->end_stream(stream_id) != 0) {
+    return NGHTTP3_ERR_CALLBACK_FAILURE;
+  }
+  return 0;
+}
+} // namespace
+
+int Http3Session::end_stream(int64_t stream_id) {
+  client_->record_ttfb();
+
+  return 0;
+}
+
 namespace {
 int recv_data(nghttp3_conn *conn, int64_t stream_id, const uint8_t *data,
               size_t datalen, void *user_data, void *stream_user_data) {
@@ -224,7 +241,7 @@ int stop_sending(nghttp3_conn *conn, int64_t stream_id, uint64_t app_error_code,
 } // namespace
 
 int Http3Session::stop_sending(int64_t stream_id, uint64_t app_error_code) {
-  auto rv = ngtcp2_conn_shutdown_stream_read(client_->quic.conn, stream_id,
+  auto rv = ngtcp2_conn_shutdown_stream_read(client_->quic.conn, 0, stream_id,
                                              app_error_code);
   if (rv != 0) {
     std::cerr << "ngtcp2_conn_shutdown_stream_read: " << ngtcp2_strerror(rv)
@@ -246,7 +263,7 @@ int reset_stream(nghttp3_conn *conn, int64_t stream_id, uint64_t app_error_code,
 } // namespace
 
 int Http3Session::reset_stream(int64_t stream_id, uint64_t app_error_code) {
-  auto rv = ngtcp2_conn_shutdown_stream_write(client_->quic.conn, stream_id,
+  auto rv = ngtcp2_conn_shutdown_stream_write(client_->quic.conn, 0, stream_id,
                                               app_error_code);
   if (rv != 0) {
     std::cerr << "ngtcp2_conn_shutdown_stream_write: " << ngtcp2_strerror(rv)
@@ -321,7 +338,7 @@ int Http3Session::init_conn() {
       h2load::recv_header,
       nullptr, // end_trailers
       h2load::stop_sending,
-      nullptr, // end_stream
+      h2load::end_stream,
       h2load::reset_stream,
       nullptr, // shutdown
   };
@@ -420,6 +437,14 @@ void Http3Session::block_stream(int64_t stream_id) {
   nghttp3_conn_block_stream(conn_, stream_id);
 }
 
+int Http3Session::unblock_stream(int64_t stream_id) {
+  if (nghttp3_conn_unblock_stream(conn_, stream_id) != 0) {
+    return -1;
+  }
+
+  return 0;
+}
+
 void Http3Session::shutdown_stream_write(int64_t stream_id) {
   nghttp3_conn_shutdown_stream_write(conn_, stream_id);
 }
