@@ -28,9 +28,9 @@
 
 #include <iostream>
 
-#ifdef HAVE_LIBNGTCP2_CRYPTO_OPENSSL
-#  include <ngtcp2/ngtcp2_crypto_openssl.h>
-#endif // HAVE_LIBNGTCP2_CRYPTO_OPENSSL
+#ifdef HAVE_LIBNGTCP2_CRYPTO_QUICTLS
+#  include <ngtcp2/ngtcp2_crypto_quictls.h>
+#endif // HAVE_LIBNGTCP2_CRYPTO_QUICTLS
 #ifdef HAVE_LIBNGTCP2_CRYPTO_BORINGSSL
 #  include <ngtcp2/ngtcp2_crypto_boringssl.h>
 #endif // HAVE_LIBNGTCP2_CRYPTO_BORINGSSL
@@ -438,11 +438,11 @@ int Client::quic_init(const sockaddr *local_addr, socklen_t local_addrlen,
       },
   };
 
-  assert(config->npn_list.size());
+  assert(config->alpn_list.size());
 
   uint32_t quic_version;
 
-  if (config->npn_list[0] == NGHTTP3_ALPN_H3) {
+  if (config->alpn_list[0] == NGHTTP3_ALPN_H3) {
     quic_version = NGTCP2_PROTO_VER_V1;
   } else {
     quic_version = NGTCP2_PROTO_VER_MIN;
@@ -538,22 +538,40 @@ void Client::quic_restart_pkt_timer() {
 }
 
 int Client::read_quic() {
-  std::array<uint8_t, 65536> buf;
+  std::array<uint8_t, 65535> buf;
   sockaddr_union su;
-  socklen_t addrlen = sizeof(su);
   int rv;
   size_t pktcnt = 0;
   ngtcp2_pkt_info pi{};
 
+  iovec msg_iov;
+  msg_iov.iov_base = buf.data();
+  msg_iov.iov_len = buf.size();
+
+  msghdr msg{};
+  msg.msg_name = &su;
+  msg.msg_iov = &msg_iov;
+  msg.msg_iovlen = 1;
+
+  uint8_t msg_ctrl[CMSG_SPACE(sizeof(uint16_t))];
+  msg.msg_control = msg_ctrl;
+
   auto ts = quic_timestamp();
 
   for (;;) {
-    auto nread =
-        recvfrom(fd, buf.data(), buf.size(), MSG_DONTWAIT, &su.sa, &addrlen);
+    msg.msg_namelen = sizeof(su);
+    msg.msg_controllen = sizeof(msg_ctrl);
+
+    auto nread = recvmsg(fd, &msg, 0);
     if (nread == -1) {
       return 0;
     }
 
+    auto gso_size = util::msghdr_get_udp_gro(&msg);
+    if (gso_size == 0) {
+      gso_size = static_cast<size_t>(nread);
+    }
+
     assert(quic.conn);
 
     ++worker->stats.udp_dgram_recv;
@@ -565,28 +583,41 @@ int Client::read_quic() {
         },
         {
             &su.sa,
-            addrlen,
+            msg.msg_namelen,
         },
     };
 
-    rv = ngtcp2_conn_read_pkt(quic.conn, &path, &pi, buf.data(), nread, ts);
-    if (rv != 0) {
-      std::cerr << "ngtcp2_conn_read_pkt: " << ngtcp2_strerror(rv) << std::endl;
+    auto data = buf.data();
 
-      if (!quic.last_error.error_code) {
-        if (rv == NGTCP2_ERR_CRYPTO) {
-          ngtcp2_ccerr_set_tls_alert(&quic.last_error,
-                                     ngtcp2_conn_get_tls_alert(quic.conn),
-                                     nullptr, 0);
-        } else {
-          ngtcp2_ccerr_set_liberr(&quic.last_error, rv, nullptr, 0);
+    for (;;) {
+      auto datalen = std::min(static_cast<size_t>(nread), gso_size);
+
+      ++pktcnt;
+
+      rv = ngtcp2_conn_read_pkt(quic.conn, &path, &pi, data, datalen, ts);
+      if (rv != 0) {
+        if (!quic.last_error.error_code) {
+          if (rv == NGTCP2_ERR_CRYPTO) {
+            ngtcp2_ccerr_set_tls_alert(&quic.last_error,
+                                       ngtcp2_conn_get_tls_alert(quic.conn),
+                                       nullptr, 0);
+          } else {
+            ngtcp2_ccerr_set_liberr(&quic.last_error, rv, nullptr, 0);
+          }
         }
+
+        return -1;
       }
 
-      return -1;
+      nread -= datalen;
+      if (nread == 0) {
+        break;
+      }
+
+      data += datalen;
     }
 
-    if (++pktcnt == 100) {
+    if (pktcnt >= 100) {
       break;
     }
   }
@@ -623,7 +654,8 @@ int Client::write_quic() {
       ngtcp2_conn_get_path_max_tx_udp_payload_size(quic.conn);
 #endif // UDP_SEGMENT
   auto max_pktcnt =
-      ngtcp2_conn_get_send_quantum(quic.conn) / max_udp_payload_size;
+      std::max(ngtcp2_conn_get_send_quantum(quic.conn) / max_udp_payload_size,
+               static_cast<size_t>(1));
   uint8_t *bufpos = quic.tx.data.get();
   ngtcp2_path_storage ps;
   size_t gso_size = 0;
