@@ -25,11 +25,11 @@
 #include "nghttp2_config.h"
 
 #ifdef __sgi
-#define daemon _daemonize
+#  define daemon _daemonize
 #endif
 
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+#  include <unistd.h>
 #endif // HAVE_UNISTD_H
 #include <signal.h>
 #include <getopt.h>
@@ -55,7 +55,7 @@ namespace nghttp2 {
 namespace {
 int parse_push_config(Config &config, const char *optarg) {
   const char *eq = strchr(optarg, '=');
-  if (eq == NULL) {
+  if (eq == nullptr) {
     return -1;
   }
   auto &paths = config.push[std::string(optarg, eq)];
@@ -63,7 +63,7 @@ int parse_push_config(Config &config, const char *optarg) {
   auto i = eq + 1;
   for (;;) {
     const char *j = strchr(i, ',');
-    if (j == NULL) {
+    if (j == nullptr) {
       j = optarg_end;
     }
     paths.emplace_back(i, j);
@@ -174,9 +174,13 @@ Options:
   --mime-types-file=<PATH>
               Path  to file  that contains  MIME media  types and  the
               extensions that represent them.
-              Default: )" << config.mime_types_file << R"(
+              Default: )"
+      << config.mime_types_file << R"(
   --no-content-length
               Don't send content-length header field.
+  --ktls      Enable ktls.
+  --no-rfc7540-pri
+              Disable RFC7540 priorities.
   --version   Display version information and exit.
   -h, --help  Display this help and exit.
 
@@ -189,12 +193,6 @@ Options:
 } // namespace
 
 int main(int argc, char **argv) {
-  tls::libssl_init();
-
-#ifndef NOTHREADS
-  tls::LibsslGlobalLock lock;
-#endif // NOTHREADS
-
   Config config;
   bool color = false;
   auto mime_types_file_set_manually = false;
@@ -227,6 +225,8 @@ int main(int argc, char **argv) {
         {"mime-types-file", required_argument, &flag, 9},
         {"no-content-length", no_argument, &flag, 10},
         {"encoder-header-table-size", required_argument, &flag, 11},
+        {"ktls", no_argument, &flag, 12},
+        {"no-rfc7540-pri", no_argument, &flag, 13},
         {nullptr, 0, nullptr, 0}};
     int option_index = 0;
     int c = getopt_long(argc, argv, "DVb:c:d:ehm:n:p:va:w:W:", long_options,
@@ -244,9 +244,15 @@ int main(int argc, char **argv) {
     case 'V':
       config.verify_client = true;
       break;
-    case 'b':
-      config.padding = strtol(optarg, nullptr, 10);
+    case 'b': {
+      auto n = util::parse_uint(optarg);
+      if (n == -1) {
+        std::cerr << "-b: Bad option value: " << optarg << std::endl;
+        exit(EXIT_FAILURE);
+      }
+      config.padding = n;
       break;
+    }
     case 'd':
       config.htdocs = optarg;
       break;
@@ -268,13 +274,12 @@ int main(int argc, char **argv) {
       std::cerr << "-n: WARNING: Threading disabled at build time, "
                 << "no threads created." << std::endl;
 #else
-      char *end;
-      errno = 0;
-      config.num_worker = strtoul(optarg, &end, 10);
-      if (errno == ERANGE || *end != '\0' || config.num_worker == 0) {
+      auto n = util::parse_uint(optarg);
+      if (n == -1) {
         std::cerr << "-n: Bad option value: " << optarg << std::endl;
         exit(EXIT_FAILURE);
       }
+      config.num_worker = n;
 #endif // NOTHREADS
       break;
     }
@@ -305,10 +310,8 @@ int main(int argc, char **argv) {
       break;
     case 'w':
     case 'W': {
-      char *endptr;
-      errno = 0;
-      auto n = strtoul(optarg, &endptr, 10);
-      if (errno != 0 || *endptr != '\0' || n >= 31) {
+      auto n = util::parse_uint(optarg);
+      if (n == -1 || n > 30) {
         std::cerr << "-" << static_cast<char>(c)
                   << ": specify the integer in the range [0, 30], inclusive"
                   << std::endl;
@@ -406,6 +409,14 @@ int main(int argc, char **argv) {
         config.encoder_header_table_size = n;
         break;
       }
+      case 12:
+        // tls option
+        config.ktls = true;
+        break;
+      case 13:
+        // no-rfc7540-pri option
+        config.no_rfc7540_pri = true;
+        break;
       }
       break;
     default:
@@ -418,7 +429,15 @@ int main(int argc, char **argv) {
     exit(EXIT_FAILURE);
   }
 
-  config.port = strtol(argv[optind++], nullptr, 10);
+  {
+    auto portStr = argv[optind++];
+    auto n = util::parse_uint(portStr);
+    if (n == -1 || n > std::numeric_limits<uint16_t>::max()) {
+      std::cerr << "<PORT>: Bad value: " << portStr << std::endl;
+      exit(EXIT_FAILURE);
+    }
+    config.port = n;
+  }
 
   if (!config.no_tls) {
     config.private_key_file = argv[optind++];
@@ -434,7 +453,7 @@ int main(int argc, char **argv) {
 #ifdef __sgi
     if (daemon(0, 0, 0, 0) == -1) {
 #else
-    if (daemon(0, 0) == -1) {
+    if (util::daemonize(0, 0) == -1) {
 #endif
       perror("daemon");
       exit(EXIT_FAILURE);
