@@ -33,6 +33,12 @@
 
 #include <openssl/ssl.h>
 
+#include <nghttp2/nghttp2.h>
+
+#ifdef ENABLE_HTTP3
+#  include <ngtcp2/ngtcp2_crypto.h>
+#endif // ENABLE_HTTP3
+
 #include "shrpx_rate_limit.h"
 #include "shrpx_error.h"
 #include "memchunk.h"
@@ -41,31 +47,47 @@ namespace shrpx {
 
 struct MemcachedRequest;
 
-enum {
-  TLS_CONN_NORMAL,
-  TLS_CONN_WAIT_FOR_SESSION_CACHE,
-  TLS_CONN_GOT_SESSION_CACHE,
-  TLS_CONN_CANCEL_SESSION_CACHE,
-  TLS_CONN_WRITE_STARTED,
+namespace tls {
+struct TLSSessionCache;
+} // namespace tls
+
+enum class TLSHandshakeState {
+  NORMAL,
+  WAIT_FOR_SESSION_CACHE,
+  GOT_SESSION_CACHE,
+  CANCEL_SESSION_CACHE,
+  WRITE_STARTED,
 };
 
 struct TLSConnection {
   DefaultMemchunks wbuf;
   DefaultPeekMemchunks rbuf;
+  // Stores TLSv1.3 early data.
+  DefaultMemchunks earlybuf;
   SSL *ssl;
   SSL_SESSION *cached_session;
   MemcachedRequest *cached_session_lookup_req;
-  ev_tstamp last_write_idle;
+  tls::TLSSessionCache *client_session_cache;
+  std::chrono::steady_clock::time_point last_write_idle;
   size_t warmup_writelen;
   // length passed to SSL_write and SSL_read last time.  This is
   // required since these functions require the exact same parameters
   // on non-blocking I/O.
   size_t last_writelen, last_readlen;
-  int handshake_state;
+  TLSHandshakeState handshake_state;
   bool initial_handshake_done;
   bool reneg_started;
   // true if ssl is prepared to do handshake as server.
   bool server_handshake;
+  // true if ssl is initialized as server, and client requested
+  // signed_certificate_timestamp extension.
+  bool sct_requested;
+  // true if TLSv1.3 early data has been completely received.  Since
+  // SSL_read_early_data acts like SSL_do_handshake, this field may be
+  // true even if the negotiated TLS version is TLSv1.2 or earlier.
+  // This value is also true if this is client side connection for
+  // convenience.
+  bool early_data_finish;
 };
 
 struct TCPHint {
@@ -84,7 +106,7 @@ struct Connection {
              const RateLimitConfig &write_limit,
              const RateLimitConfig &read_limit, IOCb writecb, IOCb readcb,
              TimerCb timeoutcb, void *data, size_t tls_dyn_rec_warmup_threshold,
-             ev_tstamp tls_dyn_rec_idle_timeout, shrpx_proto proto);
+             ev_tstamp tls_dyn_rec_idle_timeout, Proto proto);
   ~Connection();
 
   void disconnect();
@@ -93,6 +115,7 @@ struct Connection {
   void prepare_server_handshake();
 
   int tls_handshake();
+  int tls_handshake_simple();
   int write_tls_pending_handshake();
 
   int check_http2_requirement();
@@ -107,8 +130,8 @@ struct Connection {
   // underlying connection blocks), return 0.  SHRPX_ERR_EOF is
   // returned in case of EOF and no data was read.  Otherwise
   // SHRPX_ERR_NETWORK is return in case of error.
-  ssize_t write_tls(const void *data, size_t len);
-  ssize_t read_tls(void *data, size_t len);
+  nghttp2_ssize write_tls(const void *data, size_t len);
+  nghttp2_ssize read_tls(void *data, size_t len);
 
   size_t get_tls_write_limit();
   // Updates the number of bytes written in warm up period.
@@ -117,9 +140,13 @@ struct Connection {
   // determine fallback to short record size mode.
   void start_tls_write_idle();
 
-  ssize_t write_clear(const void *data, size_t len);
-  ssize_t writev_clear(struct iovec *iov, int iovcnt);
-  ssize_t read_clear(void *data, size_t len);
+  nghttp2_ssize write_clear(const void *data, size_t len);
+  nghttp2_ssize writev_clear(struct iovec *iov, int iovcnt);
+  nghttp2_ssize read_clear(void *data, size_t len);
+  // Read at most |len| bytes of data from socket without rate limit.
+  nghttp2_ssize read_nolim_clear(void *data, size_t len);
+  // Peek at most |len| bytes of data from socket without rate limit.
+  nghttp2_ssize peek_clear(void *data, size_t len);
 
   void handle_tls_pending_read();
 
@@ -133,11 +160,15 @@ struct Connection {
 
   // Restarts read timer with timeout value |t|.
   void again_rt(ev_tstamp t);
-  // Restarts read timer without chainging timeout.
+  // Restarts read timer without changing timeout.
   void again_rt();
   // Returns true if read timer expired.
   bool expired_rt();
 
+#ifdef ENABLE_HTTP3
+  // This must be the first member of Connection.
+  ngtcp2_crypto_conn_ref conn_ref;
+#endif // ENABLE_HTTP3
   TLSConnection tls;
   ev_io wev;
   ev_io rev;
@@ -149,18 +180,23 @@ struct Connection {
   void *data;
   int fd;
   size_t tls_dyn_rec_warmup_threshold;
-  ev_tstamp tls_dyn_rec_idle_timeout;
+  std::chrono::steady_clock::duration tls_dyn_rec_idle_timeout;
   // Application protocol used over the connection.  This field is not
   // used in this object at the moment.  The rest of the program may
   // use this value when it is useful.
-  shrpx_proto proto;
+  Proto proto;
   // The point of time when last read is observed.  Note: since we use
   // |rt| as idle timer, the activity is not limited to read.
-  ev_tstamp last_read;
+  std::chrono::steady_clock::time_point last_read;
   // Timeout for read timer |rt|.
   ev_tstamp read_timeout;
 };
 
+#ifdef ENABLE_HTTP3
+static_assert(std::is_standard_layout<Connection>::value,
+              "Connection is not standard layout");
+#endif // ENABLE_HTTP3
+
 // Creates BIO_method shared by all SSL objects.
 BIO_METHOD *create_bio_method();
 
