@@ -48,6 +48,7 @@
 #include "xsi_strerror.h"
 #include "util.h"
 #include "template.h"
+#include "ssl_compat.h"
 
 using namespace nghttp2;
 
@@ -264,7 +265,7 @@ int ConnectionHandler::create_single_worker() {
           nb_,
 #endif // HAVE_NEVERBLEED
           tlsconf.cacert, memcachedconf.cert_file,
-          memcachedconf.private_key_file, nullptr);
+          memcachedconf.private_key_file);
       all_ssl_ctx_.push_back(session_cache_ssl_ctx);
 #ifdef ENABLE_HTTP3
       quic_all_ssl_ctx_.push_back(nullptr);
@@ -277,15 +278,14 @@ int ConnectionHandler::create_single_worker() {
 #endif // ENABLE_HTTP3 && HAVE_LIBBPF
 
 #ifdef ENABLE_HTTP3
-  assert(cid_prefixes_.size() == 1);
-  const auto &cid_prefix = cid_prefixes_[0];
+  assert(worker_ids_.size() == 1);
+  const auto &wid = worker_ids_[0];
 #endif // ENABLE_HTTP3
 
   single_worker_ = std::make_unique<Worker>(
       loop_, sv_ssl_ctx, cl_ssl_ctx, session_cache_ssl_ctx, cert_tree_.get(),
 #ifdef ENABLE_HTTP3
-      quic_sv_ssl_ctx, quic_cert_tree_.get(), cid_prefix.data(),
-      cid_prefix.size(),
+      quic_sv_ssl_ctx, quic_cert_tree_.get(), wid,
 #  ifdef HAVE_LIBBPF
       /* index = */ 0,
 #  endif // HAVE_LIBBPF
@@ -366,7 +366,7 @@ int ConnectionHandler::create_worker_thread(size_t num) {
           nb_,
 #  endif // HAVE_NEVERBLEED
           tlsconf.cacert, memcachedconf.cert_file,
-          memcachedconf.private_key_file, nullptr);
+          memcachedconf.private_key_file);
       all_ssl_ctx_.push_back(session_cache_ssl_ctx);
 #  ifdef ENABLE_HTTP3
       quic_all_ssl_ctx_.push_back(nullptr);
@@ -375,21 +375,20 @@ int ConnectionHandler::create_worker_thread(size_t num) {
   }
 
 #  ifdef ENABLE_HTTP3
-  assert(cid_prefixes_.size() == num);
+  assert(worker_ids_.size() == num);
 #  endif // ENABLE_HTTP3
 
   for (size_t i = 0; i < num; ++i) {
     auto loop = ev_loop_new(config->ev_loop_flags);
 
 #  ifdef ENABLE_HTTP3
-    const auto &cid_prefix = cid_prefixes_[i];
+    const auto &wid = worker_ids_[i];
 #  endif // ENABLE_HTTP3
 
     auto worker = std::make_unique<Worker>(
         loop, sv_ssl_ctx, cl_ssl_ctx, session_cache_ssl_ctx, cert_tree_.get(),
 #  ifdef ENABLE_HTTP3
-        quic_sv_ssl_ctx, quic_cert_tree_.get(), cid_prefix.data(),
-        cid_prefix.size(),
+        quic_sv_ssl_ctx, quic_cert_tree_.get(), wid,
 #    ifdef HAVE_LIBBPF
         i,
 #    endif // HAVE_LIBBPF
@@ -543,9 +542,7 @@ int ConnectionHandler::handle_connection(int fd, sockaddr *addr, int addrlen,
   return 0;
 }
 
-struct ev_loop *ConnectionHandler::get_loop() const {
-  return loop_;
-}
+struct ev_loop *ConnectionHandler::get_loop() const { return loop_; }
 
 Worker *ConnectionHandler::get_single_worker() const {
   return single_worker_.get();
@@ -740,40 +737,31 @@ void ConnectionHandler::handle_ocsp_complete() {
     // that case we get nullptr.
     auto quic_ssl_ctx = quic_all_ssl_ctx_[ocsp_.next];
     if (quic_ssl_ctx) {
-#  ifndef OPENSSL_IS_BORINGSSL
       auto quic_tls_ctx_data = static_cast<tls::TLSContextData *>(
           SSL_CTX_get_app_data(quic_ssl_ctx));
-#    ifdef HAVE_ATOMIC_STD_SHARED_PTR
+#  ifdef HAVE_ATOMIC_STD_SHARED_PTR
       std::atomic_store_explicit(
           &quic_tls_ctx_data->ocsp_data,
           std::make_shared<std::vector<uint8_t>>(ocsp_.resp),
           std::memory_order_release);
-#    else  // !HAVE_ATOMIC_STD_SHARED_PTR
+#  else  // !HAVE_ATOMIC_STD_SHARED_PTR
       std::lock_guard<std::mutex> g(quic_tls_ctx_data->mu);
       quic_tls_ctx_data->ocsp_data =
           std::make_shared<std::vector<uint8_t>>(ocsp_.resp);
-#    endif // !HAVE_ATOMIC_STD_SHARED_PTR
-#  else    // OPENSSL_IS_BORINGSSL
-      SSL_CTX_set_ocsp_response(quic_ssl_ctx, ocsp_.resp.data(),
-                                ocsp_.resp.size());
-#  endif   // OPENSSL_IS_BORINGSSL
+#  endif // !HAVE_ATOMIC_STD_SHARED_PTR
     }
 #endif // ENABLE_HTTP3
 
-#ifndef OPENSSL_IS_BORINGSSL
-#  ifdef HAVE_ATOMIC_STD_SHARED_PTR
+#ifdef HAVE_ATOMIC_STD_SHARED_PTR
     std::atomic_store_explicit(
         &tls_ctx_data->ocsp_data,
         std::make_shared<std::vector<uint8_t>>(std::move(ocsp_.resp)),
         std::memory_order_release);
-#  else  // !HAVE_ATOMIC_STD_SHARED_PTR
+#else  // !HAVE_ATOMIC_STD_SHARED_PTR
     std::lock_guard<std::mutex> g(tls_ctx_data->mu);
     tls_ctx_data->ocsp_data =
         std::make_shared<std::vector<uint8_t>>(std::move(ocsp_.resp));
-#  endif // !HAVE_ATOMIC_STD_SHARED_PTR
-#else    // OPENSSL_IS_BORINGSSL
-    SSL_CTX_set_ocsp_response(ssl_ctx, ocsp_.resp.data(), ocsp_.resp.size());
-#endif   // OPENSSL_IS_BORINGSSL
+#endif // !HAVE_ATOMIC_STD_SHARED_PTR
   }
 
   ++ocsp_.next;
@@ -936,8 +924,7 @@ SSL_CTX *ConnectionHandler::create_tls_ticket_key_memcached_ssl_ctx() {
 #ifdef HAVE_NEVERBLEED
       nb_,
 #endif // HAVE_NEVERBLEED
-      tlsconf.cacert, memcachedconf.cert_file, memcachedconf.private_key_file,
-      nullptr);
+      tlsconf.cacert, memcachedconf.cert_file, memcachedconf.private_key_file);
 
   all_ssl_ctx_.push_back(ssl_ctx);
 #ifdef ENABLE_HTTP3
@@ -1019,27 +1006,23 @@ void ConnectionHandler::set_enable_acceptor_on_ocsp_completion(bool f) {
 #ifdef ENABLE_HTTP3
 int ConnectionHandler::forward_quic_packet(
     const UpstreamAddr *faddr, const Address &remote_addr,
-    const Address &local_addr, const ngtcp2_pkt_info &pi,
-    const uint8_t *cid_prefix, const uint8_t *data, size_t datalen) {
+    const Address &local_addr, const ngtcp2_pkt_info &pi, const WorkerID &wid,
+    const uint8_t *data, size_t datalen) {
   assert(!get_config()->single_thread);
 
-  for (auto &worker : workers_) {
-    if (!std::equal(cid_prefix, cid_prefix + SHRPX_QUIC_CID_PREFIXLEN,
-                    worker->get_cid_prefix())) {
-      continue;
-    }
-
-    WorkerEvent wev{};
-    wev.type = WorkerEventType::QUIC_PKT_FORWARD;
-    wev.quic_pkt = std::make_unique<QUICPacket>(faddr->index, remote_addr,
-                                                local_addr, pi, data, datalen);
-
-    worker->send(std::move(wev));
-
-    return 0;
+  auto worker = find_worker(wid);
+  if (worker == nullptr) {
+    return -1;
   }
 
-  return -1;
+  WorkerEvent wev{};
+  wev.type = WorkerEventType::QUIC_PKT_FORWARD;
+  wev.quic_pkt = std::make_unique<QUICPacket>(faddr->index, remote_addr,
+                                              local_addr, pi, data, datalen);
+
+  worker->send(std::move(wev));
+
+  return 0;
 }
 
 void ConnectionHandler::set_quic_keying_materials(
@@ -1052,22 +1035,40 @@ ConnectionHandler::get_quic_keying_materials() const {
   return quic_keying_materials_;
 }
 
-void ConnectionHandler::set_cid_prefixes(
-    const std::vector<std::array<uint8_t, SHRPX_QUIC_CID_PREFIXLEN>>
-        &cid_prefixes) {
-  cid_prefixes_ = cid_prefixes;
+void ConnectionHandler::set_worker_ids(std::vector<WorkerID> worker_ids) {
+  worker_ids_ = std::move(worker_ids);
+}
+
+namespace {
+ssize_t find_worker_index(const std::vector<WorkerID> &worker_ids,
+                          const WorkerID &wid) {
+  assert(!worker_ids.empty());
+
+  if (wid.server != worker_ids[0].server ||
+      wid.worker_process != worker_ids[0].worker_process ||
+      wid.thread >= worker_ids.size()) {
+    return -1;
+  }
+
+  return wid.thread;
+}
+} // namespace
+
+Worker *ConnectionHandler::find_worker(const WorkerID &wid) const {
+  auto idx = find_worker_index(worker_ids_, wid);
+  if (idx == -1) {
+    return nullptr;
+  }
+
+  return workers_[idx].get();
 }
 
 QUICLingeringWorkerProcess *
-ConnectionHandler::match_quic_lingering_worker_process_cid_prefix(
-    const uint8_t *dcid, size_t dcidlen) {
-  assert(dcidlen >= SHRPX_QUIC_CID_PREFIXLEN);
-
+ConnectionHandler::match_quic_lingering_worker_process_worker_id(
+    const WorkerID &wid) {
   for (auto &lwps : quic_lingering_worker_processes_) {
-    for (auto &cid_prefix : lwps.cid_prefixes) {
-      if (std::equal(std::begin(cid_prefix), std::end(cid_prefix), dcid)) {
-        return &lwps;
-      }
+    if (find_worker_index(lwps.worker_ids, wid) != -1) {
+      return &lwps;
     }
   }
 
@@ -1252,22 +1253,16 @@ int ConnectionHandler::quic_ipc_read() {
   // At the moment, UpstreamAddr index is unknown.
   pkt->upstream_addr_index = static_cast<size_t>(-1);
 
-  uint32_t version;
-  const uint8_t *dcid;
-  size_t dcidlen;
-  const uint8_t *scid;
-  size_t scidlen;
+  ngtcp2_version_cid vc;
 
-  auto rv =
-      ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                    p, datalen, SHRPX_QUIC_SCIDLEN);
+  auto rv = ngtcp2_pkt_decode_version_cid(&vc, p, datalen, SHRPX_QUIC_SCIDLEN);
   if (rv < 0) {
     LOG(ERROR) << "ngtcp2_pkt_decode_version_cid: " << ngtcp2_strerror(rv);
 
     return -1;
   }
 
-  if (dcidlen != SHRPX_QUIC_SCIDLEN) {
+  if (vc.dcidlen != SHRPX_QUIC_SCIDLEN) {
     LOG(ERROR) << "DCID length is invalid";
     return -1;
   }
@@ -1292,33 +1287,29 @@ int ConnectionHandler::quic_ipc_read() {
 
   auto &qkm = quic_keying_materials_->keying_materials.front();
 
-  std::array<uint8_t, SHRPX_QUIC_DECRYPTED_DCIDLEN> decrypted_dcid;
+  ConnectionID decrypted_dcid;
 
-  if (decrypt_quic_connection_id(decrypted_dcid.data(),
-                                 dcid + SHRPX_QUIC_CID_PREFIX_OFFSET,
-                                 qkm.cid_encryption_key.data()) != 0) {
+  if (decrypt_quic_connection_id(decrypted_dcid,
+                                 vc.dcid + SHRPX_QUIC_CID_WORKER_ID_OFFSET,
+                                 qkm.cid_decryption_ctx) != 0) {
     return -1;
   }
 
-  for (auto &worker : workers_) {
-    if (!std::equal(std::begin(decrypted_dcid),
-                    std::begin(decrypted_dcid) + SHRPX_QUIC_CID_PREFIXLEN,
-                    worker->get_cid_prefix())) {
-      continue;
+  auto worker = find_worker(decrypted_dcid.worker);
+  if (worker == nullptr) {
+    if (LOG_ENABLED(INFO)) {
+      LOG(INFO) << "No worker to match Worker ID";
     }
 
-    WorkerEvent wev{
-        .type = WorkerEventType::QUIC_PKT_FORWARD,
-        .quic_pkt = std::move(pkt),
-    };
-    worker->send(std::move(wev));
-
     return 0;
   }
 
-  if (LOG_ENABLED(INFO)) {
-    LOG(INFO) << "No worker to match CID prefix";
-  }
+  WorkerEvent wev{
+      .type = WorkerEventType::QUIC_PKT_FORWARD,
+      .quic_pkt = std::move(pkt),
+  };
+
+  worker->send(std::move(wev));
 
   return 0;
 }
