@@ -345,13 +345,26 @@ constexpr llhttp_settings_t htp_hooks = {
     nullptr,             // llhttp_cb      on_message_begin;
     nullptr,             // llhttp_data_cb on_url;
     nullptr,             // llhttp_data_cb on_status;
+    nullptr,             // llhttp_data_cb on_method;
+    nullptr,             // llhttp_data_cb on_version;
     nullptr,             // llhttp_data_cb on_header_field;
     nullptr,             // llhttp_data_cb on_header_value;
+    nullptr,             // llhttp_data_cb on_chunk_extension_name;
+    nullptr,             // llhttp_data_cb on_chunk_extension_value;
     htp_hdrs_completecb, // llhttp_cb      on_headers_complete;
     nullptr,             // llhttp_data_cb on_body;
     nullptr,             // llhttp_cb      on_message_complete;
-    nullptr,             // llhttp_cb      on_chunk_header
-    nullptr,             // llhttp_cb      on_chunk_complete
+    nullptr,             // llhttp_cb      on_url_complete;
+    nullptr,             // llhttp_cb      on_status_complete;
+    nullptr,             // llhttp_cb      on_method_complete;
+    nullptr,             // llhttp_cb      on_version_complete;
+    nullptr,             // llhttp_cb      on_header_field_complete;
+    nullptr,             // llhttp_cb      on_header_value_complete;
+    nullptr,             // llhttp_cb      on_chunk_extension_name_complete;
+    nullptr,             // llhttp_cb      on_chunk_extension_value_complete;
+    nullptr,             // llhttp_cb      on_chunk_header;
+    nullptr,             // llhttp_cb      on_chunk_complete;
+    nullptr,             // llhttp_cb      on_reset;
 };
 } // namespace
 
@@ -742,15 +755,15 @@ void Http2Session::remove_stream_data(StreamData *sd) {
 
 int Http2Session::submit_request(Http2DownstreamConnection *dconn,
                                  const nghttp2_nv *nva, size_t nvlen,
-                                 const nghttp2_data_provider *data_prd) {
+                                 const nghttp2_data_provider2 *data_prd) {
   assert(state_ == Http2SessionState::CONNECTED);
   auto sd = std::make_unique<StreamData>();
   sd->dlnext = sd->dlprev = nullptr;
   // TODO Specify nullptr to pri_spec for now
-  auto stream_id =
-      nghttp2_submit_request(session_, nullptr, nva, nvlen, data_prd, sd.get());
+  auto stream_id = nghttp2_submit_request2(session_, nullptr, nva, nvlen,
+                                           data_prd, sd.get());
   if (stream_id < 0) {
-    SSLOG(FATAL, this) << "nghttp2_submit_request() failed: "
+    SSLOG(FATAL, this) << "nghttp2_submit_request2() failed: "
                        << nghttp2_strerror(stream_id);
     return -1;
   }
@@ -1640,7 +1653,7 @@ nghttp2_session_callbacks *create_http2_downstream_callbacks() {
                                                    send_data_callback);
 
   if (get_config()->padding) {
-    nghttp2_session_callbacks_set_select_padding_callback(
+    nghttp2_session_callbacks_set_select_padding_callback2(
         callbacks, http::select_padding_callback);
   }
 
@@ -1659,14 +1672,7 @@ int Http2Session::connection_made() {
     const unsigned char *next_proto = nullptr;
     unsigned int next_proto_len = 0;
 
-#ifndef OPENSSL_NO_NEXTPROTONEG
-    SSL_get0_next_proto_negotiated(conn_.tls.ssl, &next_proto, &next_proto_len);
-#endif // !OPENSSL_NO_NEXTPROTONEG
-#if OPENSSL_VERSION_NUMBER >= 0x10002000L
-    if (!next_proto) {
-      SSL_get0_alpn_selected(conn_.tls.ssl, &next_proto, &next_proto_len);
-    }
-#endif // OPENSSL_VERSION_NUMBER >= 0x10002000L
+    SSL_get0_alpn_selected(conn_.tls.ssl, &next_proto, &next_proto_len);
 
     if (!next_proto) {
       downstream_failure(addr_, raddr_);
@@ -1693,14 +1699,17 @@ int Http2Session::connection_made() {
     return -1;
   }
 
-  std::array<nghttp2_settings_entry, 4> entry;
-  size_t nentry = 2;
+  std::array<nghttp2_settings_entry, 5> entry;
+  size_t nentry = 3;
   entry[0].settings_id = NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS;
   entry[0].value = http2conf.downstream.max_concurrent_streams;
 
   entry[1].settings_id = NGHTTP2_SETTINGS_INITIAL_WINDOW_SIZE;
   entry[1].value = http2conf.downstream.window_size;
 
+  entry[2].settings_id = NGHTTP2_SETTINGS_NO_RFC7540_PRIORITIES;
+  entry[2].value = 1;
+
   if (http2conf.no_server_push || config->http2_proxy) {
     entry[nentry].settings_id = NGHTTP2_SETTINGS_ENABLE_PUSH;
     entry[nentry].value = 0;
@@ -1745,11 +1754,9 @@ int Http2Session::on_read(const uint8_t *data, size_t datalen) {
 int Http2Session::on_write() { return on_write_(*this); }
 
 int Http2Session::downstream_read(const uint8_t *data, size_t datalen) {
-  ssize_t rv;
-
-  rv = nghttp2_session_mem_recv(session_, data, datalen);
+  auto rv = nghttp2_session_mem_recv2(session_, data, datalen);
   if (rv < 0) {
-    SSLOG(ERROR, this) << "nghttp2_session_mem_recv() returned error: "
+    SSLOG(ERROR, this) << "nghttp2_session_mem_recv2() returned error: "
                        << nghttp2_strerror(rv);
     return -1;
   }
@@ -1769,9 +1776,9 @@ int Http2Session::downstream_read(const uint8_t *data, size_t datalen) {
 int Http2Session::downstream_write() {
   for (;;) {
     const uint8_t *data;
-    auto datalen = nghttp2_session_mem_send(session_, &data);
+    auto datalen = nghttp2_session_mem_send2(session_, &data);
     if (datalen < 0) {
-      SSLOG(ERROR, this) << "nghttp2_session_mem_send() returned error: "
+      SSLOG(ERROR, this) << "nghttp2_session_mem_send2() returned error: "
                          << nghttp2_strerror(datalen);
       return -1;
     }
@@ -1818,9 +1825,7 @@ void Http2Session::signal_write() {
   }
 }
 
-struct ev_loop *Http2Session::get_loop() const {
-  return conn_.loop;
-}
+struct ev_loop *Http2Session::get_loop() const { return conn_.loop; }
 
 ev_io *Http2Session::get_wev() { return &conn_.wev; }
 
@@ -2002,7 +2007,7 @@ int Http2Session::connected() {
 }
 
 int Http2Session::read_clear() {
-  conn_.last_read = ev_now(conn_.loop);
+  conn_.last_read = std::chrono::steady_clock::now();
 
   std::array<uint8_t, 16_k> buf;
 
@@ -2024,7 +2029,7 @@ int Http2Session::read_clear() {
 }
 
 int Http2Session::write_clear() {
-  conn_.last_read = ev_now(conn_.loop);
+  conn_.last_read = std::chrono::steady_clock::now();
 
   std::array<struct iovec, MAX_WR_IOVCNT> iov;
 
@@ -2065,7 +2070,7 @@ int Http2Session::write_clear() {
 }
 
 int Http2Session::tls_handshake() {
-  conn_.last_read = ev_now(conn_.loop);
+  conn_.last_read = std::chrono::steady_clock::now();
 
   ERR_clear_error();
 
@@ -2104,7 +2109,7 @@ int Http2Session::tls_handshake() {
 }
 
 int Http2Session::read_tls() {
-  conn_.last_read = ev_now(conn_.loop);
+  conn_.last_read = std::chrono::steady_clock::now();
 
   std::array<uint8_t, 16_k> buf;
 
@@ -2128,7 +2133,7 @@ int Http2Session::read_tls() {
 }
 
 int Http2Session::write_tls() {
-  conn_.last_read = ev_now(conn_.loop);
+  conn_.last_read = std::chrono::steady_clock::now();
 
   ERR_clear_error();
 
