@@ -285,7 +285,10 @@ void Http2Upstream::on_start_request(const nghttp2_frame *frame) {
 
   downstream->reset_upstream_rtimer();
 
-  handler_->repeat_read_timer();
+  auto config = get_config();
+  auto &httpconf = config->http;
+
+  handler_->reset_upstream_read_timeout(httpconf.timeout.header);
 
   auto &req = downstream->request();
 
@@ -298,8 +301,6 @@ void Http2Upstream::on_start_request(const nghttp2_frame *frame) {
 
   ++num_requests_;
 
-  auto config = get_config();
-  auto &httpconf = config->http;
   if (httpconf.max_requests <= num_requests_) {
     start_graceful_shutdown();
   }
@@ -419,10 +420,16 @@ int Http2Upstream::on_request_headers(Downstream *downstream,
 
   downstream->set_request_state(DownstreamState::HEADER_COMPLETE);
 
+  if (config->http.require_http_scheme &&
+      !http::check_http_scheme(req.scheme, handler_->get_ssl() != nullptr)) {
+    if (error_reply(downstream, 400) != 0) {
+      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
+    }
+    return 0;
+  }
+
 #ifdef HAVE_MRUBY
-  auto upstream = downstream->get_upstream();
-  auto handler = upstream->get_client_handler();
-  auto worker = handler->get_worker();
+  auto worker = handler_->get_worker();
   auto mruby_ctx = worker->get_mruby_context();
 
   if (mruby_ctx->run_on_request_proc(downstream) != 0) {
@@ -460,7 +467,9 @@ void Http2Upstream::start_downstream(Downstream *downstream) {
 void Http2Upstream::initiate_downstream(Downstream *downstream) {
   int rv;
 
+#ifdef HAVE_MRUBY
   DownstreamConnection *dconn_ptr;
+#endif // HAVE_MRUBY
 
   for (;;) {
     auto dconn = handler_->get_downstream_connection(rv, downstream);
@@ -709,7 +718,7 @@ int on_frame_send_callback(nghttp2_session *session, const nghttp2_frame *frame,
         upstream, handler->get_mcpool(), promised_stream_id);
     auto &req = promised_downstream->request();
 
-    // As long as we use nghttp2_session_mem_send(), setting stream
+    // As long as we use nghttp2_session_mem_send2(), setting stream
     // user data here should not fail.  This is because this callback
     // is called just after frame was serialized.  So no worries about
     // hanging Downstream.
@@ -987,7 +996,7 @@ nghttp2_session_callbacks *create_http2_upstream_callbacks() {
   auto config = get_config();
 
   if (config->padding) {
-    nghttp2_session_callbacks_set_select_padding_callback(
+    nghttp2_session_callbacks_set_select_padding_callback2(
         callbacks, http::select_padding_callback);
   }
 
@@ -1037,8 +1046,8 @@ Http2Upstream::Http2Upstream(ClientHandler *handler)
   flow_control_ = true;
 
   // TODO Maybe call from outside?
-  std::array<nghttp2_settings_entry, 4> entry;
-  size_t nentry = 2;
+  std::array<nghttp2_settings_entry, 5> entry;
+  size_t nentry = 3;
 
   entry[0].settings_id = NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS;
   entry[0].value = http2conf.upstream.max_concurrent_streams;
@@ -1050,6 +1059,9 @@ Http2Upstream::Http2Upstream(ClientHandler *handler)
     entry[1].value = http2conf.upstream.window_size;
   }
 
+  entry[2].settings_id = NGHTTP2_SETTINGS_NO_RFC7540_PRIORITIES;
+  entry[2].value = 1;
+
   if (!config->http2_proxy) {
     entry[nentry].settings_id = NGHTTP2_SETTINGS_ENABLE_CONNECT_PROTOCOL;
     entry[nentry].value = 1;
@@ -1070,13 +1082,12 @@ Http2Upstream::Http2Upstream(ClientHandler *handler)
                       << nghttp2_strerror(rv);
   }
 
-  auto window_size =
-      faddr->alt_mode != UpstreamAltMode::NONE
-          ? std::numeric_limits<int32_t>::max()
-          : http2conf.upstream.optimize_window_size
-                ? std::min(http2conf.upstream.connection_window_size,
-                           NGHTTP2_INITIAL_CONNECTION_WINDOW_SIZE)
-                : http2conf.upstream.connection_window_size;
+  auto window_size = faddr->alt_mode != UpstreamAltMode::NONE
+                         ? std::numeric_limits<int32_t>::max()
+                     : http2conf.upstream.optimize_window_size
+                         ? std::min(http2conf.upstream.connection_window_size,
+                                    NGHTTP2_INITIAL_CONNECTION_WINDOW_SIZE)
+                         : http2conf.upstream.connection_window_size;
 
   rv = nghttp2_session_set_local_window_size(session_, NGHTTP2_FLAG_NONE, 0,
                                              window_size);
@@ -1122,7 +1133,7 @@ Http2Upstream::Http2Upstream(ClientHandler *handler)
 #endif // defined(TCP_INFO) && defined(TCP_NOTSENT_LOWAT)
 
   handler_->reset_upstream_read_timeout(
-      config->conn.upstream.timeout.http2_read);
+      config->conn.upstream.timeout.http2_idle);
 
   handler_->signal_write();
 }
@@ -1135,21 +1146,20 @@ Http2Upstream::~Http2Upstream() {
 }
 
 int Http2Upstream::on_read() {
-  ssize_t rv = 0;
   auto rb = handler_->get_rb();
   auto rlimit = handler_->get_rlimit();
 
   if (rb->rleft()) {
-    rv = nghttp2_session_mem_recv(session_, rb->pos(), rb->rleft());
+    auto rv = nghttp2_session_mem_recv2(session_, rb->pos(), rb->rleft());
     if (rv < 0) {
       if (rv != NGHTTP2_ERR_BAD_CLIENT_MAGIC) {
-        ULOG(ERROR, this) << "nghttp2_session_mem_recv() returned error: "
+        ULOG(ERROR, this) << "nghttp2_session_mem_recv2() returned error: "
                           << nghttp2_strerror(rv);
       }
       return -1;
     }
 
-    // nghttp2_session_mem_recv should consume all input bytes on
+    // nghttp2_session_mem_recv2 should consume all input bytes on
     // success.
     assert(static_cast<size_t>(rv) == rb->rleft());
     rb->reset();
@@ -1211,10 +1221,10 @@ int Http2Upstream::on_write() {
     }
 
     const uint8_t *data;
-    auto datalen = nghttp2_session_mem_send(session_, &data);
+    auto datalen = nghttp2_session_mem_send2(session_, &data);
 
     if (datalen < 0) {
-      ULOG(ERROR, this) << "nghttp2_session_mem_send() returned error: "
+      ULOG(ERROR, this) << "nghttp2_session_mem_send2() returned error: "
                         << nghttp2_strerror(datalen);
       return -1;
     }
@@ -1318,7 +1328,7 @@ int Http2Upstream::downstream_eof(DownstreamConnection *dconn) {
   downstream->pop_downstream_connection();
   // dconn was deleted
   dconn = nullptr;
-  // downstream wil be deleted in on_stream_close_callback.
+  // downstream will be deleted in on_stream_close_callback.
   if (downstream->get_response_state() == DownstreamState::HEADER_COMPLETE) {
     // Server may indicate the end of the request by EOF
     if (LOG_ENABLED(INFO)) {
@@ -1426,11 +1436,11 @@ int Http2Upstream::terminate_session(uint32_t error_code) {
 }
 
 namespace {
-ssize_t downstream_data_read_callback(nghttp2_session *session,
-                                      int32_t stream_id, uint8_t *buf,
-                                      size_t length, uint32_t *data_flags,
-                                      nghttp2_data_source *source,
-                                      void *user_data) {
+nghttp2_ssize downstream_data_read_callback(nghttp2_session *session,
+                                            int32_t stream_id, uint8_t *buf,
+                                            size_t length, uint32_t *data_flags,
+                                            nghttp2_data_source *source,
+                                            void *user_data) {
   int rv;
   auto downstream = static_cast<Downstream *>(source->ptr);
   auto body = downstream->get_response_buf();
@@ -1498,7 +1508,7 @@ int Http2Upstream::send_reply(Downstream *downstream, const uint8_t *body,
                               size_t bodylen) {
   int rv;
 
-  nghttp2_data_provider data_prd, *data_prd_ptr = nullptr;
+  nghttp2_data_provider2 data_prd, *data_prd_ptr = nullptr;
 
   if (bodylen) {
     data_prd.source.ptr = downstream;
@@ -1545,10 +1555,10 @@ int Http2Upstream::send_reply(Downstream *downstream, const uint8_t *body,
     nva.push_back(http2::make_nv_nocopy(p.name, p.value));
   }
 
-  rv = nghttp2_submit_response(session_, downstream->get_stream_id(),
-                               nva.data(), nva.size(), data_prd_ptr);
+  rv = nghttp2_submit_response2(session_, downstream->get_stream_id(),
+                                nva.data(), nva.size(), data_prd_ptr);
   if (nghttp2_is_fatal(rv)) {
-    ULOG(FATAL, this) << "nghttp2_submit_response() failed: "
+    ULOG(FATAL, this) << "nghttp2_submit_response2() failed: "
                       << nghttp2_strerror(rv);
     return -1;
   }
@@ -1579,7 +1589,7 @@ int Http2Upstream::error_reply(Downstream *downstream,
   body->append(html);
   downstream->set_response_state(DownstreamState::MSG_COMPLETE);
 
-  nghttp2_data_provider data_prd;
+  nghttp2_data_provider2 data_prd;
   data_prd.source.ptr = downstream;
   data_prd.read_callback = downstream_data_read_callback;
 
@@ -1597,10 +1607,10 @@ int Http2Upstream::error_reply(Downstream *downstream,
        http2::make_nv_ls_nocopy("content-length", content_length),
        http2::make_nv_ls_nocopy("date", date)}};
 
-  rv = nghttp2_submit_response(session_, downstream->get_stream_id(),
-                               nva.data(), nva.size(), &data_prd);
+  rv = nghttp2_submit_response2(session_, downstream->get_stream_id(),
+                                nva.data(), nva.size(), &data_prd);
   if (rv < NGHTTP2_ERR_FATAL) {
-    ULOG(FATAL, this) << "nghttp2_submit_response() failed: "
+    ULOG(FATAL, this) << "nghttp2_submit_response2() failed: "
                       << nghttp2_strerror(rv);
     return -1;
   }
@@ -1631,7 +1641,10 @@ void Http2Upstream::remove_downstream(Downstream *downstream) {
 
   if (downstream_queue_.get_downstreams() == nullptr) {
     // There is no downstream at the moment.  Start idle timer now.
-    handler_->repeat_read_timer();
+    auto config = get_config();
+    auto &upstreamconf = config->conn.upstream;
+
+    handler_->reset_upstream_read_timeout(upstreamconf.timeout.http2_idle);
   }
 }
 
@@ -1726,9 +1739,9 @@ int Http2Upstream::on_downstream_header_complete(Downstream *downstream) {
   }
 
   auto nva = std::vector<nghttp2_nv>();
-  // 5 means :status and possible server, via, x-http2-push, and
-  // set-cookie (for affinity cookie) header field.
-  nva.reserve(resp.fs.headers().size() + 5 +
+  // 6 means :status and possible server, via, x-http2-push, alt-svc,
+  // and set-cookie (for affinity cookie) header field.
+  nva.reserve(resp.fs.headers().size() + 6 +
               httpconf.add_response_headers.size());
 
   if (downstream->get_non_final_response()) {
@@ -1796,6 +1809,14 @@ int Http2Upstream::on_downstream_header_complete(Downstream *downstream) {
     }
   }
 
+  if (!resp.fs.header(http2::HD_ALT_SVC)) {
+    // We won't change or alter alt-svc from backend for now
+    if (!httpconf.http2_altsvc_header_value.empty()) {
+      nva.push_back(http2::make_nv_ls_nocopy(
+          "alt-svc", httpconf.http2_altsvc_header_value));
+    }
+  }
+
   auto via = resp.fs.header(http2::HD_VIA);
   if (httpconf.no_via) {
     if (via) {
@@ -1840,11 +1861,29 @@ int Http2Upstream::on_downstream_header_complete(Downstream *downstream) {
                    nva.size());
   }
 
-  nghttp2_data_provider data_prd;
+  auto priority = resp.fs.header(http2::HD_PRIORITY);
+  if (priority) {
+    nghttp2_extpri extpri;
+
+    if (nghttp2_session_get_extpri_stream_priority(
+            session_, &extpri, downstream->get_stream_id()) == 0 &&
+        nghttp2_extpri_parse_priority(&extpri, priority->value.byte(),
+                                      priority->value.size()) == 0) {
+      rv = nghttp2_session_change_extpri_stream_priority(
+          session_, downstream->get_stream_id(), &extpri,
+          /* ignore_client_signal = */ 1);
+      if (rv != 0) {
+        ULOG(ERROR, this) << "nghttp2_session_change_extpri_stream_priority: "
+                          << nghttp2_strerror(rv);
+      }
+    }
+  }
+
+  nghttp2_data_provider2 data_prd;
   data_prd.source.ptr = downstream;
   data_prd.read_callback = downstream_data_read_callback;
 
-  nghttp2_data_provider *data_prdptr;
+  nghttp2_data_provider2 *data_prdptr;
 
   if (downstream->expect_response_body() ||
       downstream->expect_response_trailer()) {
@@ -1853,10 +1892,10 @@ int Http2Upstream::on_downstream_header_complete(Downstream *downstream) {
     data_prdptr = nullptr;
   }
 
-  rv = nghttp2_submit_response(session_, downstream->get_stream_id(),
-                               nva.data(), nva.size(), data_prdptr);
+  rv = nghttp2_submit_response2(session_, downstream->get_stream_id(),
+                                nva.data(), nva.size(), data_prdptr);
   if (rv != 0) {
-    ULOG(FATAL, this) << "nghttp2_submit_response() failed";
+    ULOG(FATAL, this) << "nghttp2_submit_response2() failed";
     return -1;
   }
 
@@ -2176,7 +2215,7 @@ int Http2Upstream::submit_push_promise(const StringRef &scheme,
   // 4 for :method, :scheme, :path and :authority
   nva.reserve(4 + req.fs.headers().size());
 
-  // juse use "GET" for now
+  // just use "GET" for now
   nva.push_back(http2::make_nv_ll(":method", "GET"));
   nva.push_back(http2::make_nv_ls_nocopy(":scheme", scheme));
   nva.push_back(http2::make_nv_ls_nocopy(":path", path));
