@@ -25,7 +25,8 @@
 #include "shrpx_live_check.h"
 #include "shrpx_worker.h"
 #include "shrpx_connect_blocker.h"
-#include "shrpx_ssl.h"
+#include "shrpx_tls.h"
+#include "shrpx_log.h"
 
 namespace shrpx {
 
@@ -66,6 +67,10 @@ void timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto live_check = static_cast<LiveCheck *>(conn->data);
 
+  if (w == &conn->rt && !conn->expired_rt()) {
+    return;
+  }
+
   live_check->on_failure();
 }
 } // namespace
@@ -101,7 +106,7 @@ LiveCheck::LiveCheck(struct ev_loop *loop, SSL_CTX *ssl_ctx, Worker *worker,
             worker->get_downstream_config()->timeout.write,
             worker->get_downstream_config()->timeout.read, {}, {}, writecb,
             readcb, timeoutcb, this, get_config()->tls.dyn_rec.warmup_threshold,
-            get_config()->tls.dyn_rec.idle_timeout, PROTO_NONE),
+            get_config()->tls.dyn_rec.idle_timeout, Proto::NONE),
       wb_(worker->get_mcpool()),
       gen_(gen),
       read_(&LiveCheck::noop),
@@ -110,6 +115,7 @@ LiveCheck::LiveCheck(struct ev_loop *loop, SSL_CTX *ssl_ctx, Worker *worker,
       ssl_ctx_(ssl_ctx),
       addr_(addr),
       session_(nullptr),
+      raddr_(nullptr),
       success_count_(0),
       fail_count_(0),
       settings_ack_received_(false),
@@ -130,6 +136,16 @@ LiveCheck::~LiveCheck() {
 }
 
 void LiveCheck::disconnect() {
+  if (dns_query_) {
+    auto dns_tracker = worker_->get_dns_tracker();
+
+    dns_tracker->cancel(dns_query_.get());
+  }
+
+  dns_query_.reset();
+  // We can reuse resolved_addr_
+  raddr_ = nullptr;
+
   conn_.rlimit.stopw();
   conn_.wlimit.stopw();
 
@@ -186,42 +202,93 @@ int LiveCheck::initiate_connection() {
     return -1;
   }
 
-  if (addr_->tls) {
+  if (!dns_query_ && addr_->tls) {
     assert(ssl_ctx_);
 
-    auto ssl = ssl::create_ssl(ssl_ctx_);
+    auto ssl = tls::create_ssl(ssl_ctx_);
     if (!ssl) {
       return -1;
     }
 
     switch (addr_->proto) {
-    case PROTO_HTTP1:
-      ssl::setup_downstream_http1_alpn(ssl);
+    case Proto::HTTP1:
+      tls::setup_downstream_http1_alpn(ssl);
       break;
-    case PROTO_HTTP2:
-      ssl::setup_downstream_http2_alpn(ssl);
+    case Proto::HTTP2:
+      tls::setup_downstream_http2_alpn(ssl);
       break;
     default:
       assert(0);
     }
 
     conn_.set_ssl(ssl);
+    conn_.tls.client_session_cache = &addr_->tls_session_cache;
   }
 
-  conn_.fd = util::create_nonblock_socket(addr_->addr.su.storage.ss_family);
+  if (addr_->dns) {
+    if (!dns_query_) {
+      auto dns_query = std::make_unique<DNSQuery>(
+          addr_->host, [this](DNSResolverStatus status, const Address *result) {
+            int rv;
+
+            if (status == DNSResolverStatus::OK) {
+              *this->resolved_addr_ = *result;
+            }
+            rv = this->initiate_connection();
+            if (rv != 0) {
+              this->on_failure();
+            }
+          });
+      auto dns_tracker = worker_->get_dns_tracker();
+
+      if (!resolved_addr_) {
+        resolved_addr_ = std::make_unique<Address>();
+      }
+
+      switch (dns_tracker->resolve(resolved_addr_.get(), dns_query.get())) {
+      case DNSResolverStatus::ERROR:
+        return -1;
+      case DNSResolverStatus::RUNNING:
+        dns_query_ = std::move(dns_query);
+        return 0;
+      case DNSResolverStatus::OK:
+        break;
+      default:
+        assert(0);
+      }
+    } else {
+      switch (dns_query_->status) {
+      case DNSResolverStatus::ERROR:
+        dns_query_.reset();
+        return -1;
+      case DNSResolverStatus::OK:
+        dns_query_.reset();
+        break;
+      default:
+        assert(0);
+      }
+    }
+
+    util::set_port(*resolved_addr_, addr_->port);
+    raddr_ = resolved_addr_.get();
+  } else {
+    raddr_ = &addr_->addr;
+  }
+
+  conn_.fd = util::create_nonblock_socket(raddr_->su.storage.ss_family);
 
   if (conn_.fd == -1) {
     auto error = errno;
-    LOG(WARN) << "socket() failed; addr=" << util::to_numeric_addr(&addr_->addr)
+    LOG(WARN) << "socket() failed; addr=" << util::to_numeric_addr(raddr_)
               << ", errno=" << error;
     return -1;
   }
 
-  rv = connect(conn_.fd, &addr_->addr.su.sa, addr_->addr.len);
+  rv = connect(conn_.fd, &raddr_->su.sa, raddr_->len);
   if (rv != 0 && errno != EINPROGRESS) {
     auto error = errno;
-    LOG(WARN) << "connect() failed; addr="
-              << util::to_numeric_addr(&addr_->addr) << ", errno=" << error;
+    LOG(WARN) << "connect() failed; addr=" << util::to_numeric_addr(raddr_)
+              << ", errno=" << error;
 
     close(conn_.fd);
     conn_.fd = -1;
@@ -236,7 +303,7 @@ int LiveCheck::initiate_connection() {
       SSL_set_tlsext_host_name(conn_.tls.ssl, sni_name.c_str());
     }
 
-    auto session = ssl::reuse_tls_session(addr_->tls_session_cache);
+    auto session = tls::reuse_tls_session(addr_->tls_session_cache);
     if (session) {
       SSL_set_session(conn_.tls.ssl, session);
       SSL_SESSION_free(session);
@@ -252,17 +319,20 @@ int LiveCheck::initiate_connection() {
 
   conn_.wlimit.startw();
 
-  // TODO we should have timeout for connection establishment
+  auto &downstreamconf = *get_config()->conn.downstream;
+
+  conn_.wt.repeat = downstreamconf.timeout.connect;
   ev_timer_again(conn_.loop, &conn_.wt);
 
   return 0;
 }
 
 int LiveCheck::connected() {
-  if (!util::check_socket_connected(conn_.fd)) {
+  auto sock_error = util::get_socket_error(conn_.fd);
+  if (sock_error != 0) {
     if (LOG_ENABLED(INFO)) {
       LOG(INFO) << "Backend connect failed; addr="
-                << util::to_numeric_addr(&addr_->addr);
+                << util::to_numeric_addr(raddr_) << ": errno=" << sock_error;
     }
 
     return -1;
@@ -272,7 +342,14 @@ int LiveCheck::connected() {
     LOG(INFO) << "Connection established";
   }
 
+  auto &downstreamconf = *get_config()->conn.downstream;
+
+  // Reset timeout for write.  Previously, we set timeout for connect.
+  conn_.wt.repeat = downstreamconf.timeout.write;
+  ev_timer_again(conn_.loop, &conn_.wt);
+
   conn_.rlimit.startw();
+  conn_.again_rt();
 
   if (conn_.tls.ssl) {
     read_ = &LiveCheck::tls_handshake;
@@ -281,7 +358,7 @@ int LiveCheck::connected() {
     return do_write();
   }
 
-  if (addr_->proto == PROTO_HTTP2) {
+  if (addr_->proto == Proto::HTTP2) {
     // For HTTP/2, we try to read SETTINGS ACK from server to make
     // sure it is really alive, and serving HTTP/2.
     read_ = &LiveCheck::read_clear;
@@ -300,7 +377,7 @@ int LiveCheck::connected() {
 }
 
 int LiveCheck::tls_handshake() {
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = std::chrono::steady_clock::now();
 
   ERR_clear_error();
 
@@ -319,39 +396,26 @@ int LiveCheck::tls_handshake() {
   }
 
   if (!get_config()->tls.insecure &&
-      ssl::check_cert(conn_.tls.ssl, addr_) != 0) {
+      tls::check_cert(conn_.tls.ssl, addr_, raddr_) != 0) {
     return -1;
   }
 
-  if (!SSL_session_reused(conn_.tls.ssl)) {
-    auto tls_session = SSL_get0_session(conn_.tls.ssl);
-    if (tls_session) {
-      ssl::try_cache_tls_session(addr_->tls_session_cache, addr_->addr,
-                                 tls_session, ev_now(conn_.loop));
-    }
-  }
-
   // Check negotiated ALPN
 
   const unsigned char *next_proto = nullptr;
   unsigned int next_proto_len = 0;
 
-  SSL_get0_next_proto_negotiated(conn_.tls.ssl, &next_proto, &next_proto_len);
-#if OPENSSL_VERSION_NUMBER >= 0x10002000L
-  if (next_proto == nullptr) {
-    SSL_get0_alpn_selected(conn_.tls.ssl, &next_proto, &next_proto_len);
-  }
-#endif // OPENSSL_VERSION_NUMBER >= 0x10002000L
+  SSL_get0_alpn_selected(conn_.tls.ssl, &next_proto, &next_proto_len);
 
   auto proto = StringRef{next_proto, next_proto_len};
 
   switch (addr_->proto) {
-  case PROTO_HTTP1:
+  case Proto::HTTP1:
     if (proto.empty() || proto == StringRef::from_lit("http/1.1")) {
       break;
     }
     return -1;
-  case PROTO_HTTP2:
+  case Proto::HTTP2:
     if (util::check_h2_is_selected(proto)) {
       // For HTTP/2, we try to read SETTINGS ACK from server to make
       // sure it is really alive, and serving HTTP/2.
@@ -375,7 +439,7 @@ int LiveCheck::tls_handshake() {
 }
 
 int LiveCheck::read_tls() {
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = std::chrono::steady_clock::now();
 
   std::array<uint8_t, 4_k> buf;
 
@@ -399,6 +463,8 @@ int LiveCheck::read_tls() {
 }
 
 int LiveCheck::write_tls() {
+  conn_.last_read = std::chrono::steady_clock::now();
+
   ERR_clear_error();
 
   struct iovec iov;
@@ -406,7 +472,10 @@ int LiveCheck::write_tls() {
   for (;;) {
     if (wb_.rleft() > 0) {
       auto iovcnt = wb_.riovec(&iov, 1);
-      assert(iovcnt == 1);
+      if (iovcnt != 1) {
+        assert(0);
+        return -1;
+      }
       auto nwrite = conn_.write_tls(iov.iov_base, iov.iov_len);
 
       if (nwrite == 0) {
@@ -443,7 +512,7 @@ int LiveCheck::write_tls() {
 }
 
 int LiveCheck::read_clear() {
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = std::chrono::steady_clock::now();
 
   std::array<uint8_t, 4_k> buf;
 
@@ -465,12 +534,17 @@ int LiveCheck::read_clear() {
 }
 
 int LiveCheck::write_clear() {
+  conn_.last_read = std::chrono::steady_clock::now();
+
   struct iovec iov;
 
   for (;;) {
     if (wb_.rleft() > 0) {
       auto iovcnt = wb_.riovec(&iov, 1);
-      assert(iovcnt == 1);
+      if (iovcnt != 1) {
+        assert(0);
+        return -1;
+      }
       auto nwrite = conn_.write_clear(iov.iov_base, iov.iov_len);
 
       if (nwrite == 0) {
@@ -506,18 +580,16 @@ int LiveCheck::write_clear() {
 }
 
 int LiveCheck::on_read(const uint8_t *data, size_t len) {
-  ssize_t rv;
-
-  rv = nghttp2_session_mem_recv(session_, data, len);
+  auto rv = nghttp2_session_mem_recv2(session_, data, len);
   if (rv < 0) {
-    LOG(ERROR) << "nghttp2_session_mem_recv() returned error: "
+    LOG(ERROR) << "nghttp2_session_mem_recv2() returned error: "
                << nghttp2_strerror(rv);
     return -1;
   }
 
   if (settings_ack_received_ && !session_closing_) {
     session_closing_ = true;
-    rv = nghttp2_session_terminate_session(session_, NGHTTP2_NO_ERROR);
+    auto rv = nghttp2_session_terminate_session(session_, NGHTTP2_NO_ERROR);
     if (rv != 0) {
       return -1;
     }
@@ -545,10 +617,10 @@ int LiveCheck::on_read(const uint8_t *data, size_t len) {
 int LiveCheck::on_write() {
   for (;;) {
     const uint8_t *data;
-    auto datalen = nghttp2_session_mem_send(session_, &data);
+    auto datalen = nghttp2_session_mem_send2(session_, &data);
 
     if (datalen < 0) {
-      LOG(ERROR) << "nghttp2_session_mem_send() returned error: "
+      LOG(ERROR) << "nghttp2_session_mem_send2() returned error: "
                  << nghttp2_strerror(datalen);
       return -1;
     }
@@ -582,7 +654,7 @@ void LiveCheck::on_failure() {
   ++fail_count_;
 
   if (LOG_ENABLED(INFO)) {
-    LOG(INFO) << "Liveness check for " << util::to_numeric_addr(&addr_->addr)
+    LOG(INFO) << "Liveness check for " << addr_->host << ":" << addr_->port
               << " failed " << fail_count_ << " time(s) in a row";
   }
 
@@ -596,7 +668,7 @@ void LiveCheck::on_success() {
   fail_count_ = 0;
 
   if (LOG_ENABLED(INFO)) {
-    LOG(INFO) << "Liveness check for " << util::to_numeric_addr(&addr_->addr)
+    LOG(INFO) << "Liveness check for " << addr_->host << ":" << addr_->port
               << " succeeded " << success_count_ << " time(s) in a row";
   }
 
@@ -694,7 +766,7 @@ int LiveCheck::connection_made() {
   }
 
   auto must_terminate =
-      addr_->tls && !nghttp2::ssl::check_http2_requirement(conn_.tls.ssl);
+      addr_->tls && !nghttp2::tls::check_http2_requirement(conn_.tls.ssl);
 
   if (must_terminate) {
     if (LOG_ENABLED(INFO)) {
