@@ -34,6 +34,7 @@
 #include "shrpx_log.h"
 #include "shrpx_http3_upstream.h"
 #include "shrpx_connection_handler.h"
+#include "ssl_compat.h"
 
 namespace shrpx {
 
@@ -61,20 +62,18 @@ QUICConnectionHandler::~QUICConnectionHandler() {
 int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
                                          const Address &remote_addr,
                                          const Address &local_addr,
+                                         const ngtcp2_pkt_info &pi,
                                          const uint8_t *data, size_t datalen) {
   int rv;
-  uint32_t version;
-  const uint8_t *dcid, *scid;
-  size_t dcidlen, scidlen;
+  ngtcp2_version_cid vc;
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     data, datalen, SHRPX_QUIC_SCIDLEN);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, data, datalen, SHRPX_QUIC_SCIDLEN);
   switch (rv) {
   case 0:
     break;
   case NGTCP2_ERR_VERSION_NEGOTIATION:
-    send_version_negotiation(faddr, version, dcid, dcidlen, scid, scidlen,
-                             remote_addr, local_addr);
+    send_version_negotiation(faddr, vc.version, vc.dcid, vc.dcidlen, vc.scid,
+                             vc.scidlen, remote_addr, local_addr);
 
     return 0;
   default:
@@ -84,7 +83,7 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
   auto config = get_config();
 
   ngtcp2_cid dcid_key;
-  ngtcp2_cid_init(&dcid_key, dcid, dcidlen);
+  ngtcp2_cid_init(&dcid_key, vc.dcid, vc.dcidlen);
 
   auto conn_handler = worker_->get_connection_handler();
 
@@ -98,7 +97,7 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
     if (cwit != std::end(close_waits_)) {
       auto cw = (*cwit).second;
 
-      cw->handle_packet(faddr, remote_addr, local_addr, data, datalen);
+      cw->handle_packet(faddr, remote_addr, local_addr, pi, data, datalen);
 
       return 0;
     }
@@ -115,7 +114,7 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
         if (cwit != std::end(close_waits_)) {
           auto cw = (*cwit).second;
 
-          cw->handle_packet(faddr, remote_addr, local_addr, data, datalen);
+          cw->handle_packet(faddr, remote_addr, local_addr, pi, data, datalen);
 
           return 0;
         }
@@ -124,30 +123,29 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
   }
 
   if (it == std::end(connections_)) {
-    std::array<uint8_t, SHRPX_QUIC_DECRYPTED_DCIDLEN> decrypted_dcid;
+    ConnectionID decrypted_dcid;
 
     auto &qkms = conn_handler->get_quic_keying_materials();
     const QUICKeyingMaterial *qkm = nullptr;
 
-    if (dcidlen == SHRPX_QUIC_SCIDLEN) {
-      qkm = select_quic_keying_material(*qkms.get(), dcid);
+    if (vc.dcidlen == SHRPX_QUIC_SCIDLEN) {
+      qkm = select_quic_keying_material(
+          *qkms.get(), vc.dcid[0] & SHRPX_QUIC_DCID_KM_ID_MASK);
 
-      if (decrypt_quic_connection_id(decrypted_dcid.data(),
-                                     dcid + SHRPX_QUIC_CID_PREFIX_OFFSET,
-                                     qkm->cid_encryption_key.data()) != 0) {
+      if (decrypt_quic_connection_id(decrypted_dcid,
+                                     vc.dcid + SHRPX_QUIC_CID_WORKER_ID_OFFSET,
+                                     qkm->cid_decryption_ctx) != 0) {
         return 0;
       }
 
       if (qkm != &qkms->keying_materials.front() ||
-          !std::equal(std::begin(decrypted_dcid),
-                      std::begin(decrypted_dcid) + SHRPX_QUIC_CID_PREFIXLEN,
-                      worker_->get_cid_prefix())) {
+          decrypted_dcid.worker != worker_->get_worker_id()) {
         auto quic_lwp =
-            conn_handler->match_quic_lingering_worker_process_cid_prefix(
-                decrypted_dcid.data(), decrypted_dcid.size());
+            conn_handler->match_quic_lingering_worker_process_worker_id(
+                decrypted_dcid.worker);
         if (quic_lwp) {
           if (conn_handler->forward_quic_packet_to_lingering_worker_process(
-                  quic_lwp, remote_addr, local_addr, data, datalen) == 0) {
+                  quic_lwp, remote_addr, local_addr, pi, data, datalen) == 0) {
             return 0;
           }
 
@@ -173,40 +171,40 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
     ngtcp2_cid odcid, *podcid = nullptr;
     const uint8_t *token = nullptr;
     size_t tokenlen = 0;
+    ngtcp2_token_type token_type = NGTCP2_TOKEN_TYPE_UNKNOWN;
 
     switch (ngtcp2_accept(&hd, data, datalen)) {
     case 0: {
-      // If we get Initial and it has the CID prefix of this worker,
-      // it is likely that client is intentionally use the prefix.
-      // Just drop it.
-      if (dcidlen == SHRPX_QUIC_SCIDLEN) {
+      // If we get Initial and it has the Worker ID of this worker, it
+      // is likely that client is intentionally use the prefix.  Just
+      // drop it.
+      if (vc.dcidlen == SHRPX_QUIC_SCIDLEN) {
         if (qkm != &qkms->keying_materials.front()) {
           qkm = &qkms->keying_materials.front();
 
-          if (decrypt_quic_connection_id(decrypted_dcid.data(),
-                                         dcid + SHRPX_QUIC_CID_PREFIX_OFFSET,
-                                         qkm->cid_encryption_key.data()) != 0) {
+          if (decrypt_quic_connection_id(
+                  decrypted_dcid, vc.dcid + SHRPX_QUIC_CID_WORKER_ID_OFFSET,
+                  qkm->cid_decryption_ctx) != 0) {
             return 0;
           }
         }
 
-        if (std::equal(std::begin(decrypted_dcid),
-                       std::begin(decrypted_dcid) + SHRPX_QUIC_CID_PREFIXLEN,
-                       worker_->get_cid_prefix())) {
+        if (decrypted_dcid.worker == worker_->get_worker_id()) {
           return 0;
         }
       }
 
       if (worker_->get_graceful_shutdown()) {
-        send_connection_close(faddr, version, hd.dcid, hd.scid, remote_addr,
-                              local_addr, NGTCP2_CONNECTION_REFUSED);
+        send_connection_close(faddr, hd.version, hd.dcid, hd.scid, remote_addr,
+                              local_addr, NGTCP2_CONNECTION_REFUSED,
+                              datalen * 3);
         return 0;
       }
 
-      if (hd.token.len == 0) {
+      if (hd.tokenlen == 0) {
         if (quicconf.upstream.require_token) {
-          send_retry(faddr, version, dcid, dcidlen, scid, scidlen, remote_addr,
-                     local_addr);
+          send_retry(faddr, vc.version, vc.dcid, vc.dcidlen, vc.scid,
+                     vc.scidlen, remote_addr, local_addr, datalen * 3);
 
           return 0;
         }
@@ -214,17 +212,19 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
         break;
       }
 
-      if (dcidlen != SHRPX_QUIC_SCIDLEN) {
-        // Initial packets with token must have DCID chosen by server.
-        return 0;
-      }
+      switch (hd.token[0]) {
+      case NGTCP2_CRYPTO_TOKEN_MAGIC_RETRY: {
+        if (vc.dcidlen != SHRPX_QUIC_SCIDLEN) {
+          // Initial packets with Retry token must have DCID chosen by
+          // server.
+          return 0;
+        }
 
-      auto qkm = select_quic_keying_material(*qkms.get(), dcid);
+        auto qkm = select_quic_keying_material(
+            *qkms.get(), vc.dcid[0] & SHRPX_QUIC_DCID_KM_ID_MASK);
 
-      switch (hd.token.base[0]) {
-      case NGTCP2_CRYPTO_TOKEN_MAGIC_RETRY:
-        if (verify_retry_token(odcid, hd.token.base, hd.token.len, hd.dcid,
-                               &remote_addr.su.sa, remote_addr.len,
+        if (verify_retry_token(odcid, hd.token, hd.tokenlen, hd.version,
+                               hd.dcid, &remote_addr.su.sa, remote_addr.len,
                                qkm->secret.data(), qkm->secret.size()) != 0) {
           if (LOG_ENABLED(INFO)) {
             LOG(INFO) << "Failed to validate Retry token from remote="
@@ -233,8 +233,9 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
 
           // 2nd Retry packet is not allowed, so send CONNECTION_CLOSE
           // with INVALID_TOKEN.
-          send_connection_close(faddr, version, hd.dcid, hd.scid, remote_addr,
-                                local_addr, NGTCP2_INVALID_TOKEN);
+          send_connection_close(faddr, hd.version, hd.dcid, hd.scid,
+                                remote_addr, local_addr, NGTCP2_INVALID_TOKEN,
+                                datalen * 3);
           return 0;
         }
 
@@ -244,12 +245,39 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
         }
 
         podcid = &odcid;
-        token = hd.token.base;
-        tokenlen = hd.token.len;
+        token = hd.token;
+        tokenlen = hd.tokenlen;
+        token_type = NGTCP2_TOKEN_TYPE_RETRY;
 
         break;
-      case NGTCP2_CRYPTO_TOKEN_MAGIC_REGULAR:
-        if (verify_token(hd.token.base, hd.token.len, &remote_addr.su.sa,
+      }
+      case NGTCP2_CRYPTO_TOKEN_MAGIC_REGULAR: {
+        // If a token is a regular token, it must be at least
+        // NGTCP2_MIN_INITIAL_DCIDLEN bytes long.
+        if (vc.dcidlen < NGTCP2_MIN_INITIAL_DCIDLEN) {
+          return 0;
+        }
+
+        if (hd.tokenlen != NGTCP2_CRYPTO_MAX_REGULAR_TOKENLEN + 1) {
+          if (LOG_ENABLED(INFO)) {
+            LOG(INFO) << "Failed to validate token from remote="
+                      << util::to_numeric_addr(&remote_addr);
+          }
+
+          if (quicconf.upstream.require_token) {
+            send_retry(faddr, vc.version, vc.dcid, vc.dcidlen, vc.scid,
+                       vc.scidlen, remote_addr, local_addr, datalen * 3);
+
+            return 0;
+          }
+
+          break;
+        }
+
+        auto qkm = select_quic_keying_material(
+            *qkms.get(), hd.token[NGTCP2_CRYPTO_MAX_REGULAR_TOKENLEN]);
+
+        if (verify_token(hd.token, hd.tokenlen - 1, &remote_addr.su.sa,
                          remote_addr.len, qkm->secret.data(),
                          qkm->secret.size()) != 0) {
           if (LOG_ENABLED(INFO)) {
@@ -258,8 +286,8 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
           }
 
           if (quicconf.upstream.require_token) {
-            send_retry(faddr, version, dcid, dcidlen, scid, scidlen,
-                       remote_addr, local_addr);
+            send_retry(faddr, vc.version, vc.dcid, vc.dcidlen, vc.scid,
+                       vc.scidlen, remote_addr, local_addr, datalen * 3);
 
             return 0;
           }
@@ -272,14 +300,16 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
                     << util::to_numeric_addr(&remote_addr);
         }
 
-        token = hd.token.base;
-        tokenlen = hd.token.len;
+        token = hd.token;
+        tokenlen = hd.tokenlen;
+        token_type = NGTCP2_TOKEN_TYPE_NEW_TOKEN;
 
         break;
+      }
       default:
         if (quicconf.upstream.require_token) {
-          send_retry(faddr, version, dcid, dcidlen, scid, scidlen, remote_addr,
-                     local_addr);
+          send_retry(faddr, vc.version, vc.dcid, vc.dcidlen, vc.scid,
+                     vc.scidlen, remote_addr, local_addr, datalen * 3);
 
           return 0;
         }
@@ -289,43 +319,27 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
 
       break;
     }
-    case NGTCP2_ERR_RETRY:
-      if (worker_->get_graceful_shutdown()) {
-        send_connection_close(faddr, version, hd.dcid, hd.scid, remote_addr,
-                              local_addr, NGTCP2_CONNECTION_REFUSED);
-        return 0;
-      }
-
-      send_retry(faddr, version, dcid, dcidlen, scid, scidlen, remote_addr,
-                 local_addr);
-      return 0;
-    case NGTCP2_ERR_VERSION_NEGOTIATION:
-      send_version_negotiation(faddr, version, dcid, dcidlen, scid, scidlen,
-                               remote_addr, local_addr);
-      return 0;
     default:
-      if (!config->single_thread && !(data[0] & 0x80) &&
-          dcidlen == SHRPX_QUIC_SCIDLEN &&
-          !std::equal(std::begin(decrypted_dcid),
-                      std::begin(decrypted_dcid) + SHRPX_QUIC_CID_PREFIXLEN,
-                      worker_->get_cid_prefix())) {
-        if (conn_handler->forward_quic_packet(faddr, remote_addr, local_addr,
-                                              decrypted_dcid.data(), data,
+      if (!(data[0] & 0x80) && vc.dcidlen == SHRPX_QUIC_SCIDLEN &&
+          decrypted_dcid.worker != worker_->get_worker_id()) {
+        if (!config->single_thread &&
+            conn_handler->forward_quic_packet(faddr, remote_addr, local_addr,
+                                              pi, decrypted_dcid.worker, data,
                                               datalen) == 0) {
           return 0;
         }
-      }
 
-      if (!(data[0] & 0x80)) {
-        // TODO Must be rate limited
-        send_stateless_reset(faddr, dcid, dcidlen, remote_addr, local_addr);
+        if (datalen >= SHRPX_QUIC_SCIDLEN + 22) {
+          send_stateless_reset(faddr, datalen, vc.dcid, vc.dcidlen, remote_addr,
+                               local_addr);
+        }
       }
 
       return 0;
     }
 
     handler = handle_new_connection(faddr, remote_addr, local_addr, hd, podcid,
-                                    token, tokenlen);
+                                    token, tokenlen, token_type);
     if (handler == nullptr) {
       return 0;
     }
@@ -333,7 +347,8 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
     handler = (*it).second;
   }
 
-  if (handler->read_quic(faddr, remote_addr, local_addr, data, datalen) != 0) {
+  if (handler->read_quic(faddr, remote_addr, local_addr, pi, data, datalen) !=
+      0) {
     delete handler;
     return 0;
   }
@@ -346,7 +361,7 @@ int QUICConnectionHandler::handle_packet(const UpstreamAddr *faddr,
 ClientHandler *QUICConnectionHandler::handle_new_connection(
     const UpstreamAddr *faddr, const Address &remote_addr,
     const Address &local_addr, const ngtcp2_pkt_hd &hd, const ngtcp2_cid *odcid,
-    const uint8_t *token, size_t tokenlen) {
+    const uint8_t *token, size_t tokenlen, ngtcp2_token_type token_type) {
   std::array<char, NI_MAXHOST> host;
   std::array<char, NI_MAXSERV> service;
   int rv;
@@ -369,9 +384,9 @@ ClientHandler *QUICConnectionHandler::handle_new_connection(
     return nullptr;
   }
 
-#if OPENSSL_1_1_1_API && !defined(OPENSSL_IS_BORINGSSL)
+#ifdef NGHTTP2_GENUINE_OPENSSL
   assert(SSL_is_quic(ssl));
-#endif // OPENSSL_1_1_1_API && !defined(OPENSSL_IS_BORINGSSL)
+#endif // NGHTTP2_GENUINE_OPENSSL
 
   SSL_set_accept_state(ssl);
 
@@ -379,11 +394,11 @@ ClientHandler *QUICConnectionHandler::handle_new_connection(
   auto &quicconf = config->quic;
 
   if (quicconf.upstream.early_data) {
-#if OPENSSL_1_1_1_API && !defined(OPENSSL_IS_BORINGSSL)
+#ifdef NGHTTP2_GENUINE_OPENSSL
     SSL_set_quic_early_data_enabled(ssl, 1);
-#else  // !(OPENSSL_1_1_1_API && !defined(OPENSSL_IS_BORINGSSL))
+#elif defined(NGHTTP2_OPENSSL_IS_BORINGSSL)
     SSL_set_early_data_enabled(ssl, 1);
-#endif // !(OPENSSL_1_1_1_API && !defined(OPENSSL_IS_BORINGSSL))
+#endif // NGHTTP2_OPENSSL_IS_BORINGSSL
   }
 
   // Disable TLS session ticket if we don't have working ticket
@@ -397,8 +412,8 @@ ClientHandler *QUICConnectionHandler::handle_new_connection(
       StringRef{service.data()}, remote_addr.su.sa.sa_family, faddr);
 
   auto upstream = std::make_unique<Http3Upstream>(handler.get());
-  if (upstream->init(faddr, remote_addr, local_addr, hd, odcid, token,
-                     tokenlen) != 0) {
+  if (upstream->init(faddr, remote_addr, local_addr, hd, odcid, token, tokenlen,
+                     token_type) != 0) {
     return nullptr;
   }
 
@@ -437,7 +452,7 @@ uint32_t generate_reserved_version(const Address &addr, uint32_t version) {
 int QUICConnectionHandler::send_retry(
     const UpstreamAddr *faddr, uint32_t version, const uint8_t *ini_dcid,
     size_t ini_dcidlen, const uint8_t *ini_scid, size_t ini_scidlen,
-    const Address &remote_addr, const Address &local_addr) {
+    const Address &remote_addr, const Address &local_addr, size_t max_pktlen) {
   std::array<char, NI_MAXHOST> host;
   std::array<char, NI_MAXSERV> port;
 
@@ -456,9 +471,8 @@ int QUICConnectionHandler::send_retry(
 
   ngtcp2_cid retry_scid;
 
-  if (generate_quic_retry_connection_id(retry_scid, SHRPX_QUIC_SCIDLEN,
-                                        quicconf.server_id.data(), qkm.id,
-                                        qkm.cid_encryption_key.data()) != 0) {
+  if (generate_quic_retry_connection_id(retry_scid, quicconf.server_id, qkm.id,
+                                        qkm.cid_encryption_ctx) != 0) {
     return -1;
   }
 
@@ -469,14 +483,14 @@ int QUICConnectionHandler::send_retry(
   ngtcp2_cid_init(&idcid, ini_dcid, ini_dcidlen);
   ngtcp2_cid_init(&iscid, ini_scid, ini_scidlen);
 
-  if (generate_retry_token(token.data(), tokenlen, &remote_addr.su.sa,
+  if (generate_retry_token(token.data(), tokenlen, version, &remote_addr.su.sa,
                            remote_addr.len, retry_scid, idcid,
                            qkm.secret.data(), qkm.secret.size()) != 0) {
     return -1;
   }
 
   std::vector<uint8_t> buf;
-  buf.resize(256);
+  buf.resize(std::min(max_pktlen, static_cast<size_t>(256)));
 
   auto nwrite =
       ngtcp2_crypto_write_retry(buf.data(), buf.size(), version, &iscid,
@@ -489,8 +503,8 @@ int QUICConnectionHandler::send_retry(
   buf.resize(nwrite);
 
   quic_send_packet(faddr, &remote_addr.su.sa, remote_addr.len,
-                   &local_addr.su.sa, local_addr.len, buf.data(), buf.size(),
-                   0);
+                   &local_addr.su.sa, local_addr.len, ngtcp2_pkt_info{},
+                   buf.data(), buf.size(), 0);
 
   if (generate_quic_hashed_connection_id(idcid, remote_addr, local_addr,
                                          idcid) != 0) {
@@ -508,9 +522,7 @@ int QUICConnectionHandler::send_retry(
   auto cw = std::make_unique<CloseWait>(worker_, std::vector<ngtcp2_cid>{idcid},
                                         std::move(buf), d);
 
-  add_close_wait(cw.get());
-
-  cw.release();
+  add_close_wait(cw.release());
 
   return 0;
 }
@@ -519,10 +531,10 @@ int QUICConnectionHandler::send_version_negotiation(
     const UpstreamAddr *faddr, uint32_t version, const uint8_t *ini_dcid,
     size_t ini_dcidlen, const uint8_t *ini_scid, size_t ini_scidlen,
     const Address &remote_addr, const Address &local_addr) {
-  std::array<uint32_t, 2> sv;
-
-  sv[0] = generate_reserved_version(remote_addr, version);
-  sv[1] = NGTCP2_PROTO_VER_V1;
+  std::array<uint32_t, 2> sv{
+      generate_reserved_version(remote_addr, version),
+      NGTCP2_PROTO_VER_V1,
+  };
 
   std::array<uint8_t, NGTCP2_MAX_UDP_PAYLOAD_SIZE> buf;
 
@@ -539,15 +551,13 @@ int QUICConnectionHandler::send_version_negotiation(
   }
 
   return quic_send_packet(faddr, &remote_addr.su.sa, remote_addr.len,
-                          &local_addr.su.sa, local_addr.len, buf.data(), nwrite,
-                          0);
+                          &local_addr.su.sa, local_addr.len, ngtcp2_pkt_info{},
+                          buf.data(), nwrite, 0);
 }
 
-int QUICConnectionHandler::send_stateless_reset(const UpstreamAddr *faddr,
-                                                const uint8_t *dcid,
-                                                size_t dcidlen,
-                                                const Address &remote_addr,
-                                                const Address &local_addr) {
+int QUICConnectionHandler::send_stateless_reset(
+    const UpstreamAddr *faddr, size_t pktlen, const uint8_t *dcid,
+    size_t dcidlen, const Address &remote_addr, const Address &local_addr) {
   if (stateless_reset_bucket_ == 0) {
     if (LOG_ENABLED(INFO)) {
       LOG(INFO) << "Stateless Reset bucket has been depleted";
@@ -578,17 +588,30 @@ int QUICConnectionHandler::send_stateless_reset(const UpstreamAddr *faddr,
     return -1;
   }
 
-  std::array<uint8_t, NGTCP2_MIN_STATELESS_RESET_RANDLEN> rand_bytes;
+  // SCID + minimum expansion - NGTCP2_STATELESS_RESET_TOKENLEN
+  constexpr size_t max_rand_byteslen =
+      SHRPX_QUIC_SCIDLEN + 22 - NGTCP2_STATELESS_RESET_TOKENLEN;
 
-  if (RAND_bytes(rand_bytes.data(), rand_bytes.size()) != 1) {
+  size_t rand_byteslen;
+
+  if (pktlen <= 43) {
+    // As per
+    // https://datatracker.ietf.org/doc/html/rfc9000#section-10.3
+    rand_byteslen = pktlen - NGTCP2_STATELESS_RESET_TOKENLEN - 1;
+  } else {
+    rand_byteslen = max_rand_byteslen;
+  }
+
+  std::array<uint8_t, max_rand_byteslen> rand_bytes;
+
+  if (RAND_bytes(rand_bytes.data(), rand_byteslen) != 1) {
     return -1;
   }
 
   std::array<uint8_t, NGTCP2_MAX_UDP_PAYLOAD_SIZE> buf;
 
-  auto nwrite =
-      ngtcp2_pkt_write_stateless_reset(buf.data(), buf.size(), token.data(),
-                                       rand_bytes.data(), rand_bytes.size());
+  auto nwrite = ngtcp2_pkt_write_stateless_reset(
+      buf.data(), buf.size(), token.data(), rand_bytes.data(), rand_byteslen);
   if (nwrite < 0) {
     LOG(ERROR) << "ngtcp2_pkt_write_stateless_reset: "
                << ngtcp2_strerror(nwrite);
@@ -602,18 +625,21 @@ int QUICConnectionHandler::send_stateless_reset(const UpstreamAddr *faddr,
   }
 
   return quic_send_packet(faddr, &remote_addr.su.sa, remote_addr.len,
-                          &local_addr.su.sa, local_addr.len, buf.data(), nwrite,
-                          0);
+                          &local_addr.su.sa, local_addr.len, ngtcp2_pkt_info{},
+                          buf.data(), nwrite, 0);
 }
 
 int QUICConnectionHandler::send_connection_close(
     const UpstreamAddr *faddr, uint32_t version, const ngtcp2_cid &ini_dcid,
     const ngtcp2_cid &ini_scid, const Address &remote_addr,
-    const Address &local_addr, uint64_t error_code) {
+    const Address &local_addr, uint64_t error_code, size_t max_pktlen) {
   std::array<uint8_t, NGTCP2_MAX_UDP_PAYLOAD_SIZE> buf;
 
+  max_pktlen = std::min(max_pktlen, buf.size());
+
   auto nwrite = ngtcp2_crypto_write_connection_close(
-      buf.data(), buf.size(), version, &ini_scid, &ini_dcid, error_code);
+      buf.data(), max_pktlen, version, &ini_scid, &ini_dcid, error_code,
+      nullptr, 0);
   if (nwrite < 0) {
     LOG(ERROR) << "ngtcp2_crypto_write_connection_close failed";
     return -1;
@@ -628,8 +654,8 @@ int QUICConnectionHandler::send_connection_close(
   }
 
   return quic_send_packet(faddr, &remote_addr.su.sa, remote_addr.len,
-                          &local_addr.su.sa, local_addr.len, buf.data(), nwrite,
-                          0);
+                          &local_addr.su.sa, local_addr.len, ngtcp2_pkt_info{},
+                          buf.data(), nwrite, 0);
 }
 
 void QUICConnectionHandler::add_connection_id(const ngtcp2_cid &cid,
@@ -708,7 +734,8 @@ CloseWait::~CloseWait() {
 
 int CloseWait::handle_packet(const UpstreamAddr *faddr,
                              const Address &remote_addr,
-                             const Address &local_addr, const uint8_t *data,
+                             const Address &local_addr,
+                             const ngtcp2_pkt_info &pi, const uint8_t *data,
                              size_t datalen) {
   if (pkt.empty()) {
     return 0;
@@ -723,8 +750,8 @@ int CloseWait::handle_packet(const UpstreamAddr *faddr,
   }
 
   if (quic_send_packet(faddr, &remote_addr.su.sa, remote_addr.len,
-                       &local_addr.su.sa, local_addr.len, pkt.data(),
-                       pkt.size(), 0) != 0) {
+                       &local_addr.su.sa, local_addr.len, ngtcp2_pkt_info{},
+                       pkt.data(), pkt.size(), 0) != 0) {
     return -1;
   }
 
