@@ -26,7 +26,7 @@
 
 #include <sys/types.h>
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+#  include <unistd.h>
 #endif // HAVE_UNISTD_H
 #include <sys/resource.h>
 #include <sys/wait.h>
@@ -123,7 +123,9 @@ void graceful_shutdown(ConnectionHandler *conn_handler) {
 
   auto single_worker = conn_handler->get_single_worker();
   if (single_worker) {
-    if (single_worker->get_worker_stat()->num_connections == 0) {
+    auto worker_stat = single_worker->get_worker_stat();
+    if (worker_stat->num_connections == 0 &&
+        worker_stat->num_close_waits == 0) {
       ev_break(conn_handler->get_loop());
     }
 
@@ -178,6 +180,20 @@ void ipc_readcb(struct ev_loop *loop, ev_io *w, int revents) {
 }
 } // namespace
 
+#ifdef ENABLE_HTTP3
+namespace {
+void quic_ipc_readcb(struct ev_loop *loop, ev_io *w, int revents) {
+  auto conn_handler = static_cast<ConnectionHandler *>(w->data);
+
+  if (conn_handler->quic_ipc_read() != 0) {
+    LOG(ERROR) << "Failed to read data from QUIC IPC channel";
+
+    return;
+  }
+}
+} // namespace
+#endif // ENABLE_HTTP3
+
 namespace {
 int generate_ticket_key(TicketKey &ticket_key) {
   ticket_key.cipher = get_config()->tls.ticket.cipher;
@@ -270,14 +286,14 @@ void memcached_get_ticket_key_cb(struct ev_loop *loop, ev_timer *w,
   auto conn_handler = static_cast<ConnectionHandler *>(w->data);
   auto dispatcher = conn_handler->get_tls_ticket_key_memcached_dispatcher();
 
-  auto req = make_unique<MemcachedRequest>();
+  auto req = std::make_unique<MemcachedRequest>();
   req->key = "nghttpx:tls-ticket-key";
-  req->op = MEMCACHED_OP_GET;
+  req->op = MemcachedOp::GET;
   req->cb = [conn_handler, w](MemcachedRequest *req, MemcachedResult res) {
     switch (res.status_code) {
-    case MEMCACHED_ERR_NO_ERROR:
+    case MemcachedStatusCode::NO_ERROR:
       break;
-    case MEMCACHED_ERR_EXT_NETWORK_ERROR:
+    case MemcachedStatusCode::EXT_NETWORK_ERROR:
       conn_handler->on_tls_ticket_key_network_error(w);
       return;
     default:
@@ -382,13 +398,36 @@ void nb_child_cb(struct ev_loop *loop, ev_child *w, int revents) {
 
   ev_child_stop(loop, w);
 
-  LOG(FATAL) << "neverbleed process exitted; aborting now";
+  LOG(FATAL) << "neverbleed process exited; aborting now";
 
   nghttp2_Exit(EXIT_FAILURE);
 }
 } // namespace
 #endif // HAVE_NEVERBLEED
 
+namespace {
+int send_ready_event(int ready_ipc_fd) {
+  std::array<char, STRERROR_BUFSIZE> errbuf;
+  auto pid = getpid();
+  ssize_t nwrite;
+
+  while ((nwrite = write(ready_ipc_fd, &pid, sizeof(pid))) == -1 &&
+         errno == EINTR)
+    ;
+
+  if (nwrite < 0) {
+    auto error = errno;
+
+    LOG(ERROR) << "Writing PID to ready IPC channel failed: "
+               << xsi_strerror(error, errbuf.data(), errbuf.size());
+
+    return -1;
+  }
+
+  return 0;
+}
+} // namespace
+
 int worker_process_event_loop(WorkerProcessConfig *wpconf) {
   int rv;
   std::array<char, STRERROR_BUFSIZE> errbuf;
@@ -411,37 +450,40 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
 
   auto gen = util::make_mt19937();
 
-  ConnectionHandler conn_handler(loop, gen);
-
-  for (auto &addr : config->conn.listener.addrs) {
-    conn_handler.add_acceptor(make_unique<AcceptHandler>(&addr, &conn_handler));
-  }
-
 #ifdef HAVE_NEVERBLEED
-  {
-    std::array<char, NEVERBLEED_ERRBUF_SIZE> nb_errbuf;
-    auto nb = make_unique<neverbleed_t>();
-    if (neverbleed_init(nb.get(), nb_errbuf.data()) != 0) {
-      LOG(FATAL) << "neverbleed_init failed: " << nb_errbuf.data();
-      return -1;
-    }
-
-    LOG(NOTICE) << "neverbleed process [" << nb->daemon_pid << "] spawned";
-
-    conn_handler.set_neverbleed(std::move(nb));
+  std::array<char, NEVERBLEED_ERRBUF_SIZE> nb_errbuf;
+  auto nb = std::make_unique<neverbleed_t>();
+  if (neverbleed_init(nb.get(), nb_errbuf.data()) != 0) {
+    LOG(FATAL) << "neverbleed_init failed: " << nb_errbuf.data();
+    return -1;
   }
 
-  auto nb = conn_handler.get_neverbleed();
+  LOG(NOTICE) << "neverbleed process [" << nb->daemon_pid << "] spawned";
 
   ev_child nb_childev;
-  if (nb) {
-    ev_child_init(&nb_childev, nb_child_cb, nb->daemon_pid, 0);
-    nb_childev.data = nullptr;
-    ev_child_start(loop, &nb_childev);
-  }
 
+  ev_child_init(&nb_childev, nb_child_cb, nb->daemon_pid, 0);
+  nb_childev.data = nullptr;
+  ev_child_start(loop, &nb_childev);
 #endif // HAVE_NEVERBLEED
 
+  auto conn_handler = std::make_unique<ConnectionHandler>(loop, gen);
+
+#ifdef HAVE_NEVERBLEED
+  conn_handler->set_neverbleed(nb.get());
+#endif // HAVE_NEVERBLEED
+
+#ifdef ENABLE_HTTP3
+  conn_handler->set_quic_ipc_fd(wpconf->quic_ipc_fd);
+  conn_handler->set_quic_lingering_worker_processes(
+      wpconf->quic_lingering_worker_processes);
+#endif // ENABLE_HTTP3
+
+  for (auto &addr : config->conn.listener.addrs) {
+    conn_handler->add_acceptor(
+        std::make_unique<AcceptHandler>(&addr, conn_handler.get()));
+  }
+
   MemchunkPool mcpool;
 
   ev_timer renew_ticket_key_timer;
@@ -453,17 +495,17 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
       SSL_CTX *ssl_ctx = nullptr;
 
       if (memcachedconf.tls) {
-        ssl_ctx = conn_handler.create_tls_ticket_key_memcached_ssl_ctx();
+        ssl_ctx = conn_handler->create_tls_ticket_key_memcached_ssl_ctx();
       }
 
-      conn_handler.set_tls_ticket_key_memcached_dispatcher(
-          make_unique<MemcachedDispatcher>(
+      conn_handler->set_tls_ticket_key_memcached_dispatcher(
+          std::make_unique<MemcachedDispatcher>(
               &ticketconf.memcached.addr, loop, ssl_ctx,
               StringRef{memcachedconf.host}, &mcpool, gen));
 
       ev_timer_init(&renew_ticket_key_timer, memcached_get_ticket_key_cb, 0.,
                     0.);
-      renew_ticket_key_timer.data = &conn_handler;
+      renew_ticket_key_timer.data = conn_handler.get();
       // Get first ticket keys.
       memcached_get_ticket_key_cb(loop, &renew_ticket_key_timer, 0);
     } else {
@@ -483,14 +525,14 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
         if (!ticket_keys) {
           LOG(WARN) << "Use internal session ticket key generator";
         } else {
-          conn_handler.set_ticket_keys(std::move(ticket_keys));
+          conn_handler->set_ticket_keys(std::move(ticket_keys));
           auto_tls_ticket_key = false;
         }
       }
       if (auto_tls_ticket_key) {
         // Generate new ticket key every 1hr.
         ev_timer_init(&renew_ticket_key_timer, renew_ticket_key_cb, 0., 1_h);
-        renew_ticket_key_timer.data = &conn_handler;
+        renew_ticket_key_timer.data = conn_handler.get();
         ev_timer_again(loop, &renew_ticket_key_timer);
 
         // Generate first session ticket key before running workers.
@@ -499,8 +541,79 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
     }
   }
 
+#ifdef ENABLE_HTTP3
+  auto &quicconf = config->quic;
+
+  std::shared_ptr<QUICKeyingMaterials> qkms;
+
+  if (!quicconf.upstream.secret_file.empty()) {
+    qkms = read_quic_secret_file(quicconf.upstream.secret_file);
+    if (!qkms) {
+      LOG(WARN) << "Use QUIC keying materials generated internally";
+    }
+  }
+
+  if (!qkms) {
+    qkms = std::make_shared<QUICKeyingMaterials>();
+    qkms->keying_materials.resize(1);
+
+    auto &qkm = qkms->keying_materials.front();
+
+    if (RAND_bytes(qkm.reserved.data(), qkm.reserved.size()) != 1) {
+      LOG(ERROR) << "Failed to generate QUIC secret reserved data";
+      return -1;
+    }
+
+    if (RAND_bytes(qkm.secret.data(), qkm.secret.size()) != 1) {
+      LOG(ERROR) << "Failed to generate QUIC secret";
+      return -1;
+    }
+
+    if (RAND_bytes(qkm.salt.data(), qkm.salt.size()) != 1) {
+      LOG(ERROR) << "Failed to generate QUIC salt";
+      return -1;
+    }
+  }
+
+  for (auto &qkm : qkms->keying_materials) {
+    if (generate_quic_connection_id_encryption_key(
+            qkm.cid_encryption_key.data(), qkm.cid_encryption_key.size(),
+            qkm.secret.data(), qkm.secret.size(), qkm.salt.data(),
+            qkm.salt.size()) != 0) {
+      LOG(ERROR) << "Failed to generate QUIC Connection ID encryption key";
+      return -1;
+    }
+
+    qkm.cid_encryption_ctx = EVP_CIPHER_CTX_new();
+    if (!EVP_EncryptInit_ex(qkm.cid_encryption_ctx, EVP_aes_128_ecb(), nullptr,
+                            qkm.cid_encryption_key.data(), nullptr)) {
+      LOG(ERROR)
+          << "Failed to initialize QUIC Connection ID encryption context";
+      return -1;
+    }
+
+    EVP_CIPHER_CTX_set_padding(qkm.cid_encryption_ctx, 0);
+
+    qkm.cid_decryption_ctx = EVP_CIPHER_CTX_new();
+    if (!EVP_DecryptInit_ex(qkm.cid_decryption_ctx, EVP_aes_128_ecb(), nullptr,
+                            qkm.cid_encryption_key.data(), nullptr)) {
+      LOG(ERROR)
+          << "Failed to initialize QUIC Connection ID decryption context";
+      return -1;
+    }
+
+    EVP_CIPHER_CTX_set_padding(qkm.cid_decryption_ctx, 0);
+  }
+
+  conn_handler->set_quic_keying_materials(std::move(qkms));
+
+  conn_handler->set_worker_ids(wpconf->worker_ids);
+  conn_handler->set_quic_lingering_worker_processes(
+      wpconf->quic_lingering_worker_processes);
+#endif // ENABLE_HTTP3
+
   if (config->single_thread) {
-    rv = conn_handler.create_single_worker();
+    rv = conn_handler->create_single_worker();
     if (rv != 0) {
       return -1;
     }
@@ -518,7 +631,7 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
     }
 #endif // !NOTHREADS
 
-    rv = conn_handler.create_worker_thread(config->num_worker);
+    rv = conn_handler->create_worker_thread(config->num_worker);
     if (rv != 0) {
       return -1;
     }
@@ -533,53 +646,70 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
 #endif // !NOTHREADS
   }
 
+#if defined(ENABLE_HTTP3) && defined(HAVE_LIBBPF)
+  conn_handler->unload_bpf_objects();
+#endif // defined(ENABLE_HTTP3) && defined(HAVE_LIBBPF)
+
   drop_privileges(
 #ifdef HAVE_NEVERBLEED
-      nb
+      nb.get()
 #endif // HAVE_NEVERBLEED
   );
 
   ev_io ipcev;
   ev_io_init(&ipcev, ipc_readcb, wpconf->ipc_fd, EV_READ);
-  ipcev.data = &conn_handler;
+  ipcev.data = conn_handler.get();
   ev_io_start(loop, &ipcev);
 
+#ifdef ENABLE_HTTP3
+  ev_io quic_ipcev;
+  ev_io_init(&quic_ipcev, quic_ipc_readcb, wpconf->quic_ipc_fd, EV_READ);
+  quic_ipcev.data = conn_handler.get();
+  ev_io_start(loop, &quic_ipcev);
+#endif // ENABLE_HTTP3
+
   if (tls::upstream_tls_enabled(config->conn) && !config->tls.ocsp.disabled) {
     if (config->tls.ocsp.startup) {
-      conn_handler.set_enable_acceptor_on_ocsp_completion(true);
-      conn_handler.disable_acceptor();
+      conn_handler->set_enable_acceptor_on_ocsp_completion(true);
+      conn_handler->disable_acceptor();
     }
 
-    conn_handler.proceed_next_cert_ocsp();
+    conn_handler->proceed_next_cert_ocsp();
   }
 
   if (LOG_ENABLED(INFO)) {
     LOG(INFO) << "Entering event loop";
   }
 
+  if (send_ready_event(wpconf->ready_ipc_fd) != 0) {
+    return -1;
+  }
+
   ev_run(loop, 0);
 
-  conn_handler.cancel_ocsp_update();
+  conn_handler->cancel_ocsp_update();
+
+  // Destroy SSL_CTX held in conn_handler before killing neverbleed
+  // daemon.  Otherwise priv_rsa_finish yields "write error" and
+  // worker process aborts.
+  conn_handler.reset();
 
 #ifdef HAVE_NEVERBLEED
-  if (nb) {
-    assert(nb->daemon_pid > 0);
+  assert(nb->daemon_pid > 0);
 
-    rv = kill(nb->daemon_pid, SIGTERM);
-    if (rv != 0) {
-      auto error = errno;
-      LOG(ERROR) << "Could not send signal to neverbleed daemon: errno="
-                 << error;
-    }
+  rv = kill(nb->daemon_pid, SIGTERM);
+  if (rv != 0) {
+    auto error = errno;
+    LOG(ERROR) << "Could not send signal to neverbleed daemon: errno=" << error;
+  }
 
-    while ((rv = waitpid(nb->daemon_pid, nullptr, 0)) == -1 && errno == EINTR)
-      ;
-    if (rv == -1) {
-      auto error = errno;
-      LOG(ERROR) << "Error occurred while we were waiting for the completion "
-                    "of neverbleed process: errno="
-                 << error;
-    }
+  while ((rv = waitpid(nb->daemon_pid, nullptr, 0)) == -1 && errno == EINTR)
+    ;
+  if (rv == -1) {
+    auto error = errno;
+    LOG(ERROR) << "Error occurred while we were waiting for the completion "
+                  "of neverbleed process: errno="
+               << error;
   }
 #endif // HAVE_NEVERBLEED
 
