@@ -26,9 +26,15 @@
 #define NGHTTP2_QUEUE_TEST_H
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#  include <config.h>
 #endif /* HAVE_CONFIG_H */
 
-void test_nghttp2_queue(void);
+#define MUNIT_ENABLE_ASSERT_ALIASES
+
+#include "munit.h"
+
+extern const MunitSuite queue_suite;
+
+munit_void_test_decl(test_nghttp2_queue);
 
 #endif /* NGHTTP2_QUEUE_TEST_H */
