@@ -26,10 +26,14 @@
 #include <sys/un.h>
 #include <openssl/engine.h>
 
+#ifdef __FreeBSD__
+#include <pthread_np.h>
+#endif
+
 #ifdef __cplusplus
 extern "C" {
 #endif
-#if defined(__linux__) || defined(__FreeBSD__) || defined(__NetBSD__)
+#if (defined(__linux__) && !defined(__ANDROID__)) || defined(__FreeBSD__) || defined(__NetBSD__)
 #define NEVERBLEED_HAS_PTHREAD_SETAFFINITY_NP 1
 #if defined(__linux__)
 #define NEVERBLEED_CPU_SET_T cpu_set_t
@@ -54,6 +58,8 @@ typedef struct st_neverbleed_iobuf_t {
     char *start;
     char *end;
     size_t capacity;
+    struct st_neverbleed_iobuf_t *next;
+    unsigned processing : 1;
 } neverbleed_iobuf_t;
 
 /**
@@ -72,11 +78,20 @@ int neverbleed_setuidgid(neverbleed_t *nb, const char *user, int change_socket_o
 /**
  * builds a digestsign request
  */
-void neverbleed_start_digestsign(neverbleed_iobuf_t *buf, EVP_PKEY *pkey, const EVP_MD *md, const void *input, size_t len);
+void neverbleed_start_digestsign(neverbleed_iobuf_t *buf, EVP_PKEY *pkey, const EVP_MD *md, const void *input, size_t len,
+                                 int rsa_pss);
 /**
  * parses a digestsign response
  */
 void neverbleed_finish_digestsign(neverbleed_iobuf_t *buf, void **digest, size_t *digest_len);
+/**
+ * builds a RSA decrypt request
+ */
+void neverbleed_start_decrypt(neverbleed_iobuf_t *buf, EVP_PKEY *pkey, const void *input, size_t len);
+/**
+ * parses a decrypt response
+ */
+void neverbleed_finish_decrypt(neverbleed_iobuf_t *buf, void **digest, size_t *digest_len);
 
 #if NEVERBLEED_HAS_PTHREAD_SETAFFINITY_NP
 /**
@@ -91,18 +106,29 @@ int neverbleed_setaffinity(neverbleed_t *nb, NEVERBLEED_CPU_SET_T *cpuset);
  */
 extern void (*neverbleed_post_fork_cb)(void);
 /**
- * An optional callback used for replacing `iobuf_transaction`; i.e., the logic that sends the request and receives the response. The
- * callback returns a boolean indicating if it handled the task. It may return false to delagate the task back to the default logic.
+ * An optional callback used for replacing `iobuf_transaction`; i.e., the logic that sends the request and receives the response.
+ *
+ * If `responseless` equals `1`, the ownership of stack-allocated `req` is given to the callback. In this case, `req` must be free'd using `neverbleed_iobuf_dispose`
  */
-extern void (*neverbleed_transaction_cb)(neverbleed_iobuf_t *);
+extern void (*neverbleed_transaction_cb)(neverbleed_iobuf_t *req, int responseless);
 
 typedef void (*neverbleed_cb)(int);
 
 int neverbleed_get_fd(neverbleed_t *nb);
 static size_t neverbleed_iobuf_size(neverbleed_iobuf_t *buf);
+void neverbleed_iobuf_dispose(neverbleed_iobuf_t *buf);
 void neverbleed_transaction_read(neverbleed_t *nb, neverbleed_iobuf_t *buf);
 void neverbleed_transaction_write(neverbleed_t *nb, neverbleed_iobuf_t *buf);
 
+/**
+ * if set to a non-zero value, RSA operations are offloaded
+ */
+extern enum neverbleed_offload_type {
+    NEVERBLEED_OFFLOAD_OFF = 0,
+    NEVERBLEED_OFFLOAD_QAT_ON,
+    NEVERBLEED_OFFLOAD_QAT_AUTO,
+} neverbleed_offload;
+
 /* inline function definitions */
 
 inline size_t neverbleed_iobuf_size(neverbleed_iobuf_t *buf)
